from pytpp.attributes.user import UserAttributes
from pytpp.properties.config import IdentityAttributeValues
from pytpp.features.bases.feature_base import FeatureBase, feature
from pytpp.features.definitions.exceptions import UnexpectedValue
from pytpp.features.definitions.classes import Classes
from typing import List, Union, TYPE_CHECKING
if TYPE_CHECKING:
    from pytpp.tools.vtypes import Identity


class _IdentityBase(FeatureBase):
    def __init__(self, api):
        super().__init__(api=api)
        self._identity_dn = r'\VED\Identity'

    def _find(self, name: str, limit: int = 100, is_distribution_group: bool = False, is_security_group: bool = False,
              is_user: bool = False):
        identity_type = 0
        if is_user:
            identity_type += IdentityAttributeValues.Types.user
        if is_security_group:
            identity_type += IdentityAttributeValues.Types.security_group
        if is_distribution_group:
            identity_type += IdentityAttributeValues.Types.distribution_group

        result = self._api.websdk.Identity.Browse.post(
            filter=name,
            limit=limit,
            identity_type=identity_type
        )
        return result.identities

    def exists(self, prefixed_name: str):
        """
        Args:
            prefixed_name: The :ref:`prefixed_name` of the identity.

        Returns:
            bool: ``True`` if the identity exists, otherwise ``False``.
        """
        response = self._api.websdk.Identity.Validate.post(
            identity=self._identity_dict(prefixed_name=prefixed_name)
        )
        response.assert_valid_response()
        if response.api_response.content:
            return response.identity.prefixed_name == prefixed_name
        return False

    def get(self, prefixed_name: str = None, prefixed_universal: str = None, raise_error_if_not_exists: bool = True):
        """
        One of ``prefixed_name`` or ``prefixed_universal`` must be provided.

        Args:
            prefixed_name: The :ref:`prefixed_name` of the identity object.
            prefixed_universal: The prefixed universal of the identity object.
            raise_error_if_not_exists: Raise an exception if the identity does not exist.

        Returns:
           :ref:`identity_object`
        """
        return self._get_identity_object(
            prefixed_name=prefixed_name,
            prefixed_universal=prefixed_universal,
            raise_error_if_not_exists=raise_error_if_not_exists
        )

    def get_memberships(self, identity: 'Union[Identity.Identity, str]'):
        """
        Args:
            identity: :ref:`identity_object` or :ref:`prefixed_name` of the identity.

        Returns:
            List of :ref:`identity_object` for each group.
        """
        if isinstance(identity, str):
            identity = self._get_identity_object(prefixed_name=identity)
        memberships = self._api.websdk.Identity.GetMemberships.post(
            identity=self._identity_dict(
                prefixed_name=identity.prefixed_name,
                prefixed_universal=identity.prefixed_universal
            )
        ).identities

        return memberships

    def read_attribute(self, identity: 'Union[Identity.Identity, str]', attribute_name: str):
        """
        Args:
            identity: :ref:`identity_object` or :ref:`prefixed_name` of the identity.
            attribute_name: The name of the attribute.

        Returns:
            List[str]: List of attribute values.
        """
        prefixed_name = self._get_prefixed_name(identity)
        result = self._api.websdk.Identity.ReadAttribute.post(
            identity=self._identity_dict(prefixed_name=prefixed_name),
            attribute_name=attribute_name
        )
        return result.attributes


@feature('User')
class User(_IdentityBase):
    def __init__(self, api):
        super().__init__(api=api)

    def create(self, name: str, password: str, email_address: str, first_name: str = None, last_name: str = None,
               add_to_everyone_group: bool = True, get_if_already_exists: bool = True):
        """
        Args:
            name: Name of the user. The *local:* prefix is not required.
            password: Password.
            email_address: E-mail address.
            first_name: First name of the user.
            last_name: Last name of the user.
            add_to_everyone_group: Add the user to the ``local:Everyone`` group.
            get_if_already_exists: If the identity already exists, just return it as is.

        Returns:
            :ref:`identity_object` of the user.
        """
        attributes = {
            UserAttributes.internet_email_address: email_address,
            UserAttributes.given_name: first_name,
            UserAttributes.surname: last_name
        }
        user = self._config_create(
            name=name,
            parent_folder_dn=self._identity_dn,
            config_class=Classes.user,
            attributes=attributes,
            get_if_already_exists=get_if_already_exists
        )
        user = self.set_password(user=f'local:{user.name}', new_password=password)

        if add_to_everyone_group:
            response = self._api.websdk.Identity.AddGroupMembers.put(
                group=self._identity_dict(prefixed_name='local:Everyone'),
                members=[self._identity_dict(prefixed_name=user.prefixed_name)]
            )
            response.assert_valid_response()
        return user

    def delete(self, user: 'Union[Identity.Identity, str]'):
        """
        Deletes the user from the Local Identity Provider. The user is removed from all local groups.

        Args:
            user: :ref:`identity_object` or :ref:`prefixed_name` of the user.
        """
        if isinstance(user, str):
            user = self._get_identity_object(user)
        groups = self.get_memberships(identity=user)
        for group in groups:
            result = self._api.websdk.Identity.RemoveGroupMembers.put(
                group=self._identity_dict(prefixed_name=group.prefixed_name),
                members=[self._identity_dict(prefixed_name=user.prefixed_name)]
            )
            result.assert_valid_response()
        self._config_delete(object_dn=f'{self._identity_dn}\\{user.name}')

    def find(self, name: str, limit: int = 100):
        """
        Finds users within the the Identity Providers to which the authenticated user has view permissions.
        
        Args:
            name: String of characters contained within the names to be found.
            limit: Maximum number of results to return.

        Returns:
            List of :ref:`identity_object` for each user found.
        """
        return self._find(
            name=name,
            limit=limit,
            is_distribution_group=False,
            is_security_group=False,
            is_user=True
        )

    def set_password(self, user: 'Union[Identity.Identity, str]', new_password: str, old_password: str = None):
        """
        Sets the ``new_password`` for a local user. If the user did not have a previous password, then
        the ``old_password`` is not required.

        Args:
            user: :ref:`identity_object` or :ref:`prefixed_name` of the user.
            new_password: The new password for the user.
            old_password: The old password for the user. Required only if it exists.

        Returns:
            :ref:`identity_object` of the user.
        """
        prefixed_name = self._get_prefixed_name(user)
        response = self._api.websdk.Identity.SetPassword.post(
            identity=self._identity_dict(prefixed_name=prefixed_name),
            old_password=old_password,
            password=new_password
        )
        return response.identity


@feature('Group')
class Group(_IdentityBase):
    def __init__(self, api):
        super().__init__(api=api)

    def add_members(self, group: 'Union[Identity.Identity, str]', members: 'List[Union[Identity.Identity, str]]'):
        """
        Args:
            group: :ref:`identity_object` or :ref:`prefixed_name` of the group.
            members: List of :ref:`identity_object` or :ref:`prefixed_name` of each member.

        Returns:
            List of :ref:`identity_object` for each member.
        """
        member_prefixed_names = [self._get_prefixed_name(i) for i in members]
        prefixed_name = self._get_prefixed_name(group)
        result = self._api.websdk.Identity.AddGroupMembers.put(
            group=self._identity_dict(prefixed_name=prefixed_name),
            members=[
                self._identity_dict(prefixed_name=member_prefixed_name)
                for member_prefixed_name in member_prefixed_names
            ],
            show_members=True
        )

        if member_prefixed_names and result.invalid_members:
            im = "\n\t".join([m.prefixed_name for m in result.invalid_members])
            raise UnexpectedValue(
                f'Unable to add these members to the group "{prefixed_name}":\n\t{im}'
            )

        return result.members

    def create(self, name: str, members: 'List[Union[Identity, Identity, str]]' = None, get_if_already_exists: bool = True):
        """
        Args:
            name: Name of the user. The *local:* prefix is not required.
            members: List of :ref:`identity_object` or :ref:`prefixed_name` of each member.
            get_if_already_exists: If the identity already exists, just return it as is.

        Returns:
            :ref:`identity_object` of the group.
        """
        member_prefixed_names = [self._get_prefixed_name(i) for i in members] if members else []
        if not name.startswith('local:'):
            name = f'local:{name}'
        if get_if_already_exists:
            if self.exists(prefixed_name=name):
                return self.get(prefixed_name=name)

        result = self._api.websdk.Identity.AddGroup.post(
            name=self._identity_dict(prefixed_name=name),
            members=[
                self._identity_dict(prefixed_name=member_prefixed_name)
                for member_prefixed_name in member_prefixed_names
            ]
        )

        if member_prefixed_names and result.invalid_members:
            im = "\n\t".join([m.prefixed_name for m in result.invalid_members])
            self._log_warning_message(
                msg=f'Unable to add these members to the group "{name}":\n\t{im}'
            )

        group = result.identity
        return group

    def delete(self, group: 'Union[Identity.Identity, str]'):
        """
        Deletes a group, but not its members. All group permissions and privileges are deleted.

        Args:
            group: :ref:`identity_object` or :ref:`prefixed_name` of the group.
        """
        if isinstance(group, str):
            group = self._get_identity_object(group)
        result = self._api.websdk.Identity.Group.Prefix(group.prefix).Principal(group.universal).delete()
        result.assert_valid_response()

    def find(self, name: str, limit: int = 100, is_distribution_group: bool = False, is_security_group: bool = True):
        """
        Finds groups within the the Identity Providers to which the authenticated user has view permissions.

        Args:
            name: String of characters contained within the names to be found.
            limit: Maximum number of results to return.
            is_distribution_group: If ``True``, results include AD Distribution Groups.
            is_security_group: If ``True``, results include Local and AD/LDAP groups.

        Returns:
            List of :ref:`identity_object` for each group found.
        """
        return self._find(
            name=name,
            limit=limit,
            is_distribution_group=is_distribution_group,
            is_security_group=is_security_group,
            is_user=False
        )

    def get_members(self, group: 'Union[Identity.Identity, str]', resolve_nested: bool = False):
        """
        Args:
            group: :ref:`identity_object` or :ref:`prefixed_name` of the group.
            resolve_nested: If ``True``, recursively returns members of groups within this group.

        Returns:
            List of :ref:`identity_object` for each member of the group.
        """
        prefixed_name = self._get_prefixed_name(group)
        result = self._api.websdk.Identity.GetMembers.post(
            identity=self._identity_dict(prefixed_name=prefixed_name),
            resolve_nested=int(resolve_nested)
        )

        return result.identities

    def remove_members(self, group: 'Union[Identity.Identity, str]', members: 'List[Union[Identity.Identity, str]]'):
        """
        Removes members from a local group.

        Args:
            group: :ref:`identity_object` or :ref:`prefixed_name` of the group.
            members: List of :ref:`identity_object` or :ref:`prefixed_name` of each member.

        Returns:
            List of :ref:`identity_object` for each remaining member. If no members remain, then
            the list is empty.
        """
        member_prefixed_names = [self._get_prefixed_name(i) for i in members]
        prefixed_name = self._get_prefixed_name(group)
        result = self._api.websdk.Identity.RemoveGroupMembers.put(
            group=self._identity_dict(prefixed_name=prefixed_name),
            members=[
                self._identity_dict(prefixed_name=member_prefixed_name)
                for member_prefixed_name in member_prefixed_names
            ],
            show_members=True
        )

        return result.members

    def rename(self, group: 'Union[Identity.Identity, str]', new_group_name: str):
        """
        Renames a local group. The *local:* prefix is not required.

        Args:
            group: :ref:`identity_object` or :ref:`prefixed_name` of the group.
            new_group_name: New name of the group. No prefix required.

        Returns:
            :ref:`identity_object` with the new group name.
        """
        prefixed_name = self._get_prefixed_name(group)
        if not new_group_name.startswith('local:'):
            new_group_name = new_group_name.lstrip('local:')

        result = self._api.websdk.Identity.RenameGroup.put(
            group=self._identity_dict(prefixed_name=prefixed_name),
            new_group_name=new_group_name
        )

        return result.identity



from pytpp.attributes._helper import IterableMeta, Attribute


class ReportFilterBaseAttributes(metaclass=IterableMeta):
	__config_class__ = "Report Filter Base"
	address = Attribute('Address')
	discoverydn = Attribute('DiscoveryDN')
	grouping = Attribute('Grouping')
	longrunning = Attribute('LongRunning', min_version='16.4')
	policydn = Attribute('PolicyDN')
	reporton = Attribute('ReportOn')
	selectedcontacts = Attribute('SelectedContacts')



from pytpp.attributes._helper import IterableMeta, Attribute
from pytpp.attributes.application_base import ApplicationBaseAttributes


class ImpervaMXAttributes(ApplicationBaseAttributes, metaclass=IterableMeta):
	__config_class__ = "Imperva MX"
	file_validation_disabled = Attribute('File Validation Disabled', min_version='15.3')
	key_store_validation_disabled = Attribute('Key Store Validation Disabled')
	private_key_name = Attribute('Private Key Name')
	server_group = Attribute('Server Group')
	service = Attribute('Service')
	site = Attribute('Site')
	username_credential = Attribute('Username Credential')
	utility_path = Attribute('Utility Path')



from pytpp.attributes._helper import IterableMeta, Attribute
from pytpp.attributes.agent_base import AgentBaseAttributes
from pytpp.attributes.driver_base import DriverBaseAttributes
from pytpp.attributes.legacy_key_base import LegacyKeyBaseAttributes
from pytpp.attributes.monitoring_base import MonitoringBaseAttributes


class SSHKeyAttributes(AgentBaseAttributes, DriverBaseAttributes, LegacyKeyBaseAttributes, MonitoringBaseAttributes, metaclass=IterableMeta):
	__config_class__ = "SSH Key"
	public_key_vault_id = Attribute('Public Key Vault Id')



from pytpp.attributes._helper import IterableMeta, Attribute
from pytpp.attributes.ssh_key import SSHKeyAttributes
from pytpp.attributes.validation_base import ValidationBaseAttributes


class SSHServerKeyAttributes(SSHKeyAttributes, ValidationBaseAttributes, metaclass=IterableMeta):
	__config_class__ = "SSH Server Key"
	network_validation_disabled = Attribute('Network Validation Disabled')



from pytpp.api.api_base import API, APIResponse, api_response_property
from pytpp.properties.response_objects.preferences import Preferences


class _Preferences(API):
    def __init__(self, api_obj):
        super().__init__(api_obj=api_obj, url='/Preferences')

    def get(self, category: str = None, name: str = None, product: str = None):
        params = {
            'Category': category,
            'Name': name,
            'Product': product
        }

        class _Response(APIResponse):
            def __init__(self, response):
                super().__init__(response=response)

            @property
            @api_response_property()
            def preferences(self):
                return [Preferences.Preference(p) for p in self._from_json(key='Preferences')]

        return _Response(response=self._get(params=params))

    def post(self, preferences: dict):
        body = {
            'Preferences': preferences
        }

        return APIResponse(response=self._post(data=body))

    def delete(self, category: str = None, name: str = None, product: str = None):
        params = {
            'Category': category,
            'Name': name,
            'Product': product
        }

        return APIResponse(response=self._delete(params=params))



from pytpp.attributes._helper import IterableMeta, Attribute
from pytpp.attributes.application_base import ApplicationBaseAttributes


class Layer7SSGAttributes(ApplicationBaseAttributes, metaclass=IterableMeta):
	__config_class__ = "Layer 7 SSG"
	certificate_oid = Attribute('Certificate OID')
	certificate_only = Attribute('Certificate Only')
	chain_cert = Attribute('Chain Cert')
	database_validation_disabled = Attribute('Database Validation Disabled')
	file_validation_disabled = Attribute('File Validation Disabled', min_version='15.3')
	first_backup_certificate_oid = Attribute('First Backup Certificate OID')
	first_backup_private_key_oid = Attribute('First Backup Private Key OID')
	https_port = Attribute('HTTPS Port')
	network_validation_disabled = Attribute('Network Validation Disabled')
	outbound_ssl = Attribute('Outbound SSL')
	private_key_oid = Attribute('Private Key OID')
	provision_manner = Attribute('Provision Manner')
	revocation_checking_enabled = Attribute('Revocation Checking Enabled')
	saml_attesting = Attribute('SAML Attesting')
	second_backup_certificate_oid = Attribute('Second Backup Certificate OID')
	second_backup_private_key_oid = Attribute('Second Backup Private Key OID')
	service = Attribute('Service')
	signing_client = Attribute('Signing Client')
	signing_outbound_ssl = Attribute('Signing Outbound SSL')
	signing_saml = Attribute('Signing SAML')
	trust_anchor = Attribute('Trust Anchor')
	verify_hostname = Attribute('Verify Hostname')
	web_credential = Attribute('Web Credential')
	web_service_url = Attribute('Web Service URL')



class ConfigAttributeDataTypes:
    data_types = {
         0: 'None',
         1: 'Base64',
         2: 'Boolean',
         3: 'CaseExactString',
         4: 'CaseIgnoreList',
         5: 'CaseIgnoreString, String',
         6: 'Counter',
         7: 'DistinguishedName',
         8: 'EMailAddress',
         9: 'FaxNumber',
        10: 'Hold',
        11: 'Integer',
        12: 'Interval',
        13: 'IPNetworkAddress',
        14: 'NetworkAddress',
        15: 'NumericString',
        16: 'OctetList',
        17: 'OctetString',
        18: 'Path',
        19: 'PostalAddress',
        20: 'PrintableString',
        21: 'SchemaClass',
        22: 'Stream',
        23: 'TelephoneNumber',
        24: 'Time',
    }


from typing import List
from pytpp.api.api_base import API, APIResponse, api_response_property
from pytpp.properties.response_objects.identity import Identity


class _Teams(API):
    def __init__(self, api_obj):
        super().__init__(api_obj=api_obj, url='/Teams')
        self.AddTeamMembers = self._AddTeamMembers(api_obj=api_obj)
        self.AddTeamOwners = self._AddTeamOwners(api_obj=api_obj)
        self.DemoteTeamOwners = self._DemoteTeamOwners(api_obj=api_obj)
        self.RemoveTeamMembers = self._RemoveTeamMembers(api_obj=api_obj)
        self.RenameTeam = self._RenameTeam(api_obj=api_obj)

    def post(self, name: str, owners: list, assets: list = None, description: str = None, members: list = None,
             products: list = None):
        body = {
            'Assets': assets,
            'Description': description,
            'Name': name,
            'Owners': owners,
            'Members': members,
            'Products': products
        }

        class _Response(APIResponse):
            def __init__(self, response):
                super().__init__(response=response)

            @property
            @api_response_property()
            def identity(self):
                return [Identity.Identity(owner) for owner in self._from_json(key='Owners')]

            @property
            @api_response_property()
            def invalid_members(self):
                return [Identity.InvalidIdentity(member) for member in self._from_json(key='InvalidMembers')]

        return _Response(response=self._post(data=body))

    class _AddTeamMembers(API):
        def __init__(self, api_obj):
            super().__init__(api_obj=api_obj, url='/Teams/AddTeamMembers')

        def put(self, members: list, team: dict = None, show_members: bool = None):
            body = {
                'Members': members,
                'Team': team,
                'ShowMembers': show_members
            }

            class _Response(APIResponse):
                def __init__(self, response):
                    super().__init__(response=response)

                @property
                @api_response_property()
                def invalid_members(self):
                    return [Identity.InvalidIdentity(member) for member in self._from_json(key='InvalidMembers')]

                @property
                @api_response_property()
                def members(self):
                    return [Identity.Identity(member) for member in self._from_json(key='Members')]

                @property
                @api_response_property()
                def message(self) -> str:
                    return self._from_json(key='Message')

            return _Response(response=self._put(data=body))

    class _AddTeamOwners(API):
        def __init__(self, api_obj):
            super().__init__(api_obj=api_obj, url='/Teams/AddTeamOwners')

        def put(self, owners: list = None, team: dict = None, show_members: bool = None):
            body = {
                'Owners': owners,
                'Team': team,
                'ShowMembers': show_members
            }

            class _Response(APIResponse):
                def __init__(self, response):
                    super().__init__(response=response)

                @property
                @api_response_property()
                def owners(self):
                    return [Identity.Identity(owner) for owner in self._from_json(key='Owners')]

                @property
                @api_response_property()
                def members(self):
                    return [Identity.Identity(member) for member in self._from_json(key='Members')]

                @property
                @api_response_property()
                def message(self) -> str:
                    return self._from_json(key='Message')

            return _Response(response=self._put(data=body))

    class _DemoteTeamOwners(API):
        def __init__(self, api_obj):
            super().__init__(api_obj=api_obj, url='/Teams/DemoteTeamOwners')

        def put(self, owners: list = None, team: dict = None, show_members: bool = None):
            body = {
                'Owners': owners,
                'Team': team,
                'ShowMembers': show_members
            }

            class _Response(APIResponse):
                def __init__(self, response):
                    super().__init__(response=response)

                @property
                @api_response_property()
                def invalid_owners(self):
                    return [Identity.InvalidIdentity(owner) for owner in self._from_json(key='InvalidOwners')]

                @property
                @api_response_property()
                def owners(self):
                    return [Identity.Identity(owner) for owner in self._from_json(key='Owners')]

                @property
                @api_response_property()
                def members(self):
                    return [Identity.Identity(member) for member in self._from_json(key='Members')]

                @property
                @api_response_property()
                def message(self) -> str:
                    return self._from_json(key='Message')

            return _Response(response=self._put(data=body))

    def Prefix(self, prefix='local'):
        return self._Prefix(prefix=prefix, api_obj=self._api_obj)

    class _Prefix:
        def __init__(self, prefix: str, api_obj):
            self._prefix = prefix
            self._api_obj = api_obj

        def Universal(self, universal):
            return self._Universal(
                prefix=self._prefix,
                universal=universal,
                api_obj=self._api_obj
            )

        class _Universal(API):
            def __init__(self, prefix: str, universal: str, api_obj):
                super().__init__(api_obj=api_obj, url=f'/Teams/{prefix}/{universal}')

            def delete(self):
                
                class _Response(APIResponse):
                    def __init__(self, response):
                        super().__init__(response=response)

                return _Response(response=self._delete())

            def get(self):
                class _Response(APIResponse):
                    def __init__(self, response):
                        super().__init__(response=response)

                    @property
                    @api_response_property()
                    def assets(self) -> List[str]:
                        return self._from_json(key='Assets')

                    @property
                    @api_response_property()
                    def description(self) -> str:
                        return self._from_json(key='Description')

                    @property
                    @api_response_property()
                    def identity(self):
                        return Identity.Identity(self._from_json(key='ID'))

                    @property
                    @api_response_property()
                    def members(self):
                        return [Identity.Identity(member) for member in self._from_json(key='Members')]

                    @property
                    @api_response_property()
                    def owners(self):
                        return [Identity.Identity(owner) for owner in self._from_json(key='Owners')]

                    @property
                    @api_response_property()
                    def products(self) -> List[str]:
                        return self._from_json(key='Products')

                return _Response(response=self._get())

            def put(self, assets: list, description: str, name: str, owners: list, members: list, products: list):
                body = {
                    'Assets': assets,
                    'Description': description,
                    'Name': name,
                    'Owners': owners,
                    'Members': members,
                    'Products': products
                }

                class _Response(APIResponse):
                    def __init__(self, response):
                        super().__init__(response=response)

                    @property
                    @api_response_property()
                    def identity(self):
                        return Identity.Identity(self._from_json(key='ID'))

                    @property
                    @api_response_property()
                    def invalid_owners(self):
                        return [Identity.InvalidIdentity(owner) for owner in self._from_json(key='InvalidOwners')]

                    @property
                    @api_response_property()
                    def invalid_members(self):
                        return [Identity.InvalidIdentity(member) for member in self._from_json(key='InvalidMembers')]

                    @property
                    @api_response_property()
                    def message(self) -> str:
                        return self._from_json(key='Message')

                return _Response(response=self._put(data=body))

    class _RemoveTeamMembers(API):
        def __init__(self, api_obj):
            super().__init__(api_obj=api_obj, url='/Teams/RemoveTeamMembers')

        def put(self, team: dict, members: list = None, show_members: bool = None):
            body = {
                'Team': team,
                'Members': members,
                'ShowMembers': show_members
            }

            class _Response(APIResponse):
                def __init__(self, response):
                    super().__init__(response=response)

                @property
                @api_response_property()
                def invalid_members(self):
                    return [Identity.InvalidIdentity(member) for member in self._from_json(key='InvalidMembers')]

                @property
                @api_response_property()
                def members(self):
                    return [Identity.Identity(member) for member in self._from_json(key='Members')]

                @property
                @api_response_property()
                def message(self) -> str:
                    return self._from_json(key='Message')

                @property
                @api_response_property()
                def owners(self):
                    return [Identity.Identity(owner) for owner in self._from_json(key='Owners')]

            return _Response(response=self._put(data=body))

    class _RenameTeam(API):
        def __init__(self, api_obj):
            super().__init__(api_obj=api_obj, url='Teams/RenameTeam')

        def put(self, team: str, new_team_name: str):
            body = {
                'Team'       : team,
                'NewTeamName': new_team_name
            }

            class _Response(APIResponse):
                def __init__(self, response):
                    super().__init__(response=response)

                @property
                @api_response_property()
                def identity(self):
                    return Identity.Identity(self._from_json(key='ID'))

                @property
                @api_response_property()
                def message(self) -> str:
                    return self._from_json(key='Message')

            return _Response(response=self._put(data=body))



from pytpp.attributes._helper import IterableMeta, Attribute
from pytpp.attributes.certificate_authority_base import CertificateAuthorityBaseAttributes


class OpenSSLCAAttributes(CertificateAuthorityBaseAttributes, metaclass=IterableMeta):
	__config_class__ = "OpenSSL CA"
	certificate_directory = Attribute('Certificate Directory')
	certificate_file = Attribute('Certificate File')
	configuration_file = Attribute('Configuration File')
	copy_extensions = Attribute('Copy Extensions')
	private_key_file = Attribute('Private Key File')
	private_key_password_credential = Attribute('Private Key Password Credential')
	temp_directory = Attribute('Temp Directory')



from dataclasses import dataclass
from datetime import datetime
from typing import List


@dataclass
class Result:
    code: int
    client_result: str


@dataclass
class Client:
    client_id: int
    client_type: str
    fqdn: str
    os_name: str
    username: str


@dataclass
class ClientDetails:
    certificate_device: str
    client_id: int
    client_type: str
    client_version: str
    created_on: datetime
    dns_name: str
    effective_work: list
    fqdn: str
    groups: list
    host_domain: str
    hostname: str
    last_seen_on: datetime
    networks: 'List[_Network]'
    os_build: str
    os_name: str
    os_service_pack: str
    os_version: str
    region: str
    serial_number: str
    ssh_device: str
    system_architecture: str
    system_chassis: str
    system_manufacturer: str
    system_model: str
    trust_level: str
    username: str
    virtual_machine_id: str


@dataclass
class Work:
    associated_groups: list
    work_dn: str
    work_name: str
    work_type: str


@dataclass
class _Network:
    ip_address: str
    mac_address: str



import datetime
import time
from typing import List, Union, TYPE_CHECKING
from pytpp.features.bases.feature_base import FeatureBase, feature
from pytpp.features.definitions.exceptions import InvalidResultCode, UnexpectedValue, InvalidFormat
from pytpp.properties.config import ClientWorkAttributeValues
from pytpp.attributes.client_work_base import ClientWorkBaseAttributes
from pytpp.attributes.client_agent_configuration_work import ClientAgentConfigurationWorkAttributes
from pytpp.attributes.client_agent_automatic_upgrade_work import ClientAgentAutomaticUpgradeWorkAttributes
from pytpp.attributes.server_agent_cert_device_placement_work import ServerAgentCertDevicePlacementWorkAttributes
from pytpp.attributes.client_certificate_discovery_work import ClientCertificateDiscoveryWorkAttributes
from pytpp.attributes.network_device_certificate_work import NetworkDeviceCertificateWorkAttributes
from pytpp.attributes.certificate_provisioning_work import CertificateProvisioningWorkAttributes
from pytpp.attributes.client_certificate_work import ClientCertificateWorkAttributes
from pytpp.attributes.server_certificate_work import ServerCertificateWorkAttributes
from pytpp.attributes.server_agent_ssh_device_placement_work import ServerAgentSSHDevicePlacementWorkAttributes
from pytpp.attributes.client_agent_ssh_discovery_work import ClientAgentSSHDiscoveryWorkAttributes
from pytpp.attributes.client_agent_ssh_key_usage_work import ClientAgentSSHKeyUsageWorkAttributes
from pytpp.attributes.client_agent_ssh_provisioning_work import ClientAgentSSHProvisioningWorkAttributes
from pytpp.attributes.client_user_certificate_work import ClientUserCertificateWorkAttributes
if TYPE_CHECKING:
    from pytpp.tools.vtypes import Config, Identity


class _ClientWorkBase(FeatureBase):
    def __init__(self, api):
        super().__init__(api)
        self._work_base_dn = r'\VED\Clients\Work'

    def delete(self, work: 'Union[Config.Object, str]'):
        """
        Deletes the client work.

        Args:
            work: :ref:`config_object` or name of the client work.
        """
        work_dn = self._get_dn(work, parent_dn=self._work_base_dn)
        response = self._api.websdk.Config.Delete.post(work_dn)

        if response.result.code != 1:
            raise InvalidResultCode(
                code=response.result.code,
                code_description=response.result.credential_result
            )

    def disable(self, work: 'Union[Config.Object, str]'):
        """
        Disables the client work.

        Args:
            work: :ref:`config_object` or name of the client work.
        """
        work_dn = self._get_dn(work, parent_dn=self._work_base_dn)
        result = self._api.websdk.Config.Write.post(
            object_dn=work_dn,
            attribute_data=self._name_value_list({
                ClientWorkBaseAttributes.disabled: ["1"]
            }, keep_list_values=True)
        ).result

        if result.code != 1:
            raise InvalidResultCode(code=result.code, code_description=result.config_result)

    def enable(self, work: 'Union[Config.Object, str]'):
        """
        Enables the client work.

        Args:
            work: :ref:`config_object` or name of the client work.
        """
        work_dn = self._get_dn(work, parent_dn=self._work_base_dn)
        result = self._api.websdk.Config.ClearAttribute.post(
            object_dn=work_dn,
            attribute_name=ClientWorkBaseAttributes.disabled
        ).result

        if result.code != 1:
            raise InvalidResultCode(code=result.code, code_description=result.config_result)

    def get(self, name: str, raise_error_if_not_exists: bool = True):
        """
        Args:
            name: The name of the client work.
            raise_error_if_not_exists: Raise an exception if the client work does not exist.

        Returns:
            :ref:`config_object` of the client work.
        """
        return self._get_config_object(
            object_dn=rf'{self._work_base_dn}\{name}',
            raise_error_if_not_exists=raise_error_if_not_exists
        )

    def list(self):
        """
        Lists all client work.

        Returns:
            List of :ref:`config_object` of all client work.
        """
        response = self._api.websdk.Config.Enumerate.post(object_dn=self._work_base_dn)

        if response.result.code != 1:
            raise InvalidResultCode(code=response.result.code,
                                                 code_description=response.result.credential_result)
        return response.objects


@feature('Agent Connectivity')
class AgentConnectivity(_ClientWorkBase):
    def __init__(self, api):
        super().__init__(api=api)

    def schedule(self, work: 'Union[Config.Object, str]', start_time: int = None, daily: bool = False, hourly: bool = False,
                 days_of_week: List[str] = None, days_of_month: List[str] = None, randomize_minutes: int = 0):
        """
        .. note::
            Only one of daily, hourly, days_of_week or days_of_month can be set.

        Schedules the Agent Connectivity work to run.

        Args:
            work: :ref:`config_object` or name of the client work.
            start_time: The 24-hour UTC hour format (i.e. 20 = 8PM UTC) for the job to start.
            daily: Runs the client work daily
            hourly: Runs the client work hourly
            days_of_week: Runs the client work on specific days of the week. It is a Zero-based index of the days of the week (i.e. Sunday = '0').
            days_of_month: Runs the client work on specific days of the month.
            randomize_minutes: Randomize the given minutes for agent check-in to the server
        """
        work_dn = self._get_dn(work, parent_dn=self._work_base_dn)

        attributes = {
            ClientAgentConfigurationWorkAttributes.start_time: datetime.time(start_time % 24).strftime("%I:00 %p"),
            ClientAgentConfigurationWorkAttributes.interval  : randomize_minutes
        }

        if len([x for x in [daily, hourly, days_of_week, days_of_month] if x not in[None, False]]) != 1:
            raise InvalidFormat(
                "Error in Schedule: must specify one (and only one) of: daily,hourly,days_of_week,days_of_month")

        if daily:
            if not start_time:
                raise InvalidFormat(
                    "Error in Schedule: must supply a 'start_time' to use the daily interval")

            attributes[ClientAgentConfigurationWorkAttributes.schedule_type] = ClientWorkAttributeValues.AgentConnectivity.ScheduleType.daily
        elif hourly:
            attributes[ClientAgentConfigurationWorkAttributes.schedule_type] = ClientWorkAttributeValues.AgentConnectivity.ScheduleType.hourly
        elif days_of_week:
            if not start_time:
                raise InvalidFormat(
                    "Error in Schedule: must supply a 'start_time' to use the days_of_week interval")

            attributes[ClientAgentConfigurationWorkAttributes.schedule_type] = ClientWorkAttributeValues.AgentConnectivity.ScheduleType.days_of_week
            attributes[ClientAgentConfigurationWorkAttributes.days_of_week] = days_of_week
        elif days_of_month:
            if not start_time:
                raise InvalidFormat(
                    "Error in Schedule: must supply a 'start_time' to use the days_of_month interval")

            attributes[ClientAgentConfigurationWorkAttributes.schedule_type] = ClientWorkAttributeValues.AgentConnectivity.ScheduleType.days_of_month
            attributes[ClientAgentConfigurationWorkAttributes.days_of_month] = days_of_month
        else:
            raise InvalidFormat(
                "Error in Schedule: must supply at one of (daily, hourly, days_of_week, days_of_month)")

        response = self._api.websdk.Config.Write.post(
            object_dn=work_dn,
            attribute_data=self._name_value_list(attributes, True)
        )
        response.assert_valid_response()

    def create(self, name: str, server_url: str = "", proxy_url: str = "", proxy_credentials: 'Union[Config.Object, str]' = None,
               log_threshold: str = ClientWorkAttributeValues.AgentConnectivity.LogThreshold.info,
               attributes: dict = None, get_if_already_exists: bool = True):
        """
        Args:
            name: The name of the client work.
            server_url: The server url.
            proxy_url: The proxy url.
            proxy_credentials: :ref:`config_object` or :ref:`dn` of the proxy credentials.
            log_threshold: The log threshold.
            attributes: Additional attributes to apply to the object.
            get_if_already_exists: If the objects already exists, just return it as is.

        Returns:
            :ref:`config_object` of the client work.
        """
        work_attributes = {
            ClientAgentConfigurationWorkAttributes.created_by   : ClientWorkAttributeValues.AgentConnectivity.CreatedBy.websdk,
            ClientAgentConfigurationWorkAttributes.interval     : 0,
            ClientAgentConfigurationWorkAttributes.log_threshold: log_threshold
        }

        if len(server_url) > 0:
            work_attributes[ClientAgentConfigurationWorkAttributes.web_service_url] = server_url
        if len(proxy_url) > 0:
            work_attributes[ClientAgentConfigurationWorkAttributes.proxy_host] = proxy_url
        if len(proxy_credentials) > 0:
            work_attributes[ClientAgentConfigurationWorkAttributes.proxy_credential] = self._get_dn(proxy_credentials)

        if attributes:
            work_attributes.update(attributes)

        return self._config_create(
            name=name,
            parent_folder_dn=self._work_base_dn,
            config_class=ClientAgentConfigurationWorkAttributes.__config_class__,
            attributes=work_attributes,
            get_if_already_exists=get_if_already_exists
        )

    def unschedule(self, work: 'Union[Config.Object, str]'):
        """
        Removes any scheduling for the client work, but does not delete the client work.

        Args:
            work: :ref:`config_object` or name of the client work.
        """
        work_dn = self._get_dn(work, parent_dn=self._work_base_dn)
        for attribute_name in {
            ClientAgentConfigurationWorkAttributes.start_time,
            ClientAgentConfigurationWorkAttributes.schedule_type,
            ClientAgentConfigurationWorkAttributes.interval,
            ClientAgentConfigurationWorkAttributes.days_of_week,
            ClientAgentConfigurationWorkAttributes.days_of_month
        }:
            self._api.websdk.Config.ClearAttribute.post(
                object_dn=work_dn,
                attribute_name=attribute_name
            ).assert_valid_response()


@feature('Agent Upgrade')
class AgentUpgrade(_ClientWorkBase):
    def __init__(self, api):
        super().__init__(api=api)

    def create(self, name: str, attributes: dict = None, get_if_already_exists: bool = True):
        """
        Args:
            name: The name of the client work.
            attributes: Additional attributes to apply to the object.
            get_if_already_exists: If the objects already exists, just return it as is.

        Returns:
            :ref:`config_object` of the client work.
        """
        work_attributes = {
            ClientAgentAutomaticUpgradeWorkAttributes.created_by: ClientWorkAttributeValues.AgentUpgrade.CreatedBy.websdk,
        }

        if attributes:
            work_attributes.update(attributes)

        return self._config_create(
            name=name,
            parent_folder_dn=self._work_base_dn,
            config_class=ClientAgentAutomaticUpgradeWorkAttributes.__config_class__,
            attributes=work_attributes,
            get_if_already_exists=get_if_already_exists
        )


@feature('Certificate Device Placement')
class CertificateDevicePlacement(_ClientWorkBase):
    def __init__(self, api):
        super().__init__(api=api)

    def create(self, name: str, placement_folder: 'Union[Config.Object, str]', share_mode: int = 2,
               attributes: dict = None, get_if_already_exists: bool = True):
        """
        Args:
            name: The name of the client work.
            placement_folder: :ref:`config_object` or :ref:`dn` of the folder to place devices.
            share_mode: specify how newly discovered devices are de-duplicated
                0: search the entire policy tree
                1: search the devices folder
                2: search the devices folder and any sub-folders
                3: create a duplicate device
            attributes: Additional attributes to apply to the object.
            get_if_already_exists: If the objects already exists, just return it as is.

        Returns:
            :ref:`config_object` of the client work.
        """
        placement_folder_dn = self._get_dn(placement_folder)
        work_attributes = {
            ServerAgentCertDevicePlacementWorkAttributes.created_by            :
                ClientWorkAttributeValues.CertificateDevicePlacement.CreatedBy.websdk,
            ServerAgentCertDevicePlacementWorkAttributes.device_object_location:
                placement_folder_dn
        }

        if share_mode == 0:
            work_attributes[ServerAgentCertDevicePlacementWorkAttributes.device_share_mode] = \
                ClientWorkAttributeValues.CertificateDevicePlacement.DeviceSharedMode.whole_tree
        elif share_mode == 1:
            work_attributes[
                ServerAgentCertDevicePlacementWorkAttributes.device_share_mode] = \
                ClientWorkAttributeValues.CertificateDevicePlacement.DeviceSharedMode.devices_folder
        elif share_mode == 2:
            work_attributes[ServerAgentCertDevicePlacementWorkAttributes.device_share_mode] = \
                ClientWorkAttributeValues.CertificateDevicePlacement.DeviceSharedMode.devices_folder_and_sub_folders
        elif share_mode == 3:
            work_attributes[ServerAgentCertDevicePlacementWorkAttributes.device_share_mode] = \
                ClientWorkAttributeValues.CertificateDevicePlacement.DeviceSharedMode.duplicate_device
        else:
            raise UnexpectedValue(f"Unexpected value for 'share_mode': {share_mode}")

        if attributes:
            work_attributes.update(attributes)

        return self._config_create(
            name=name,
            parent_folder_dn=self._work_base_dn,
            config_class=ServerAgentCertDevicePlacementWorkAttributes.__config_class__,
            attributes=work_attributes,
            get_if_already_exists=get_if_already_exists
        )


@feature('Certificate Discovery')
class CertificateDiscovery(_ClientWorkBase):
    def __init__(self, api):
        super().__init__(api=api)

    def schedule(self, work: 'Union[Config.Object, str]', start_time: int = None, daily: bool = False, hourly: bool = False,
                 on_receipt: bool = False,
                 days_of_week: List[str] = None, days_of_month: List[str] = None, randomize_minutes: int = 0,
                 full_scan: bool = False):
        """
        .. note::
			Only one of daily, hourly, on_receipt, days_of_week or days_of_month can be set.

        Schedules the Certificate Discovery work to run.

        Args:
            work: :ref:`config_object` or name of the client work.
            start_time: The 24-hour UTC hour format (i.e. 20 = 8PM UTC) for the job to start.
            daily: Runs the client work daily
            hourly: Runs the client work hourly
            on_receipt: Runs the client work on_receipt
            days_of_week: Runs the client work on specific days of the week. It is a Zero-based index of the days of the week (i.e. Sunday = '0').
            days_of_month: Runs the client work on specific days of the month.
            randomize_minutes: Randomize the given minutes for agent to send data back to the server
            full_scan: Reset the cache and perform a full scan (resend all the data to the server)
        """
        work_dn = self._get_dn(work, parent_dn=self._work_base_dn)
        attributes = {
            ClientCertificateDiscoveryWorkAttributes.start_time: datetime.time(start_time % 24).strftime("%I:00 %p"),
            ClientCertificateDiscoveryWorkAttributes.interval  : randomize_minutes
        }

        if len([x for x in [daily, hourly, on_receipt, days_of_week, days_of_month] if x not in [None, False]]) != 1:
            raise InvalidFormat(
                "Error in Schedule: must specify one (and only one) of: daily,hourly,on_receipt,days_of_week,days_of_month")

        if full_scan:
            attributes[ClientCertificateDiscoveryWorkAttributes.clear_cache_timestamp] = str(time.time())

        if daily:
            if not start_time:
                raise InvalidFormat(
                    "Error in Schedule: must supply a 'start_time' to use the daily interval")

            attributes[ClientCertificateDiscoveryWorkAttributes.schedule_type] = \
                ClientWorkAttributeValues.CertificateDiscovery.ScheduleType.daily
        elif hourly:
            attributes[ClientCertificateDiscoveryWorkAttributes.schedule_type] = \
                ClientWorkAttributeValues.CertificateDiscovery.ScheduleType.hourly
        elif on_receipt:
            attributes[ClientCertificateDiscoveryWorkAttributes.schedule_type] = \
                ClientWorkAttributeValues.CertificateDiscovery.ScheduleType.on_receipt
        elif days_of_week:
            if not start_time:
                raise InvalidFormat(
                    "Error in Schedule: must supply a 'start_time' to use the days_of_week interval")

            attributes[ClientCertificateDiscoveryWorkAttributes.schedule_type] = \
                ClientWorkAttributeValues.CertificateDiscovery.ScheduleType.days_of_week
            attributes[ClientCertificateDiscoveryWorkAttributes.days_of_week] = days_of_week
        elif days_of_month:
            if not start_time:
                raise InvalidFormat(
                    "Error in Schedule: must supply a 'start_time' to use the days_of_month interval")

            attributes[ClientCertificateDiscoveryWorkAttributes.schedule_type] = \
                ClientWorkAttributeValues.CertificateDiscovery.ScheduleType.days_of_month
            attributes[ClientCertificateDiscoveryWorkAttributes.days_of_month] = days_of_month
        else:
            raise InvalidFormat(
                "Error in Schedule: must supply at one of (daily, hourly, days_of_week, days_of_month)")

        response = self._api.websdk.Config.Write.post(
            object_dn=work_dn,
            attribute_data=self._name_value_list(attributes, True)
        )
        response.assert_valid_response()

    def create(self,
               name: str, certificate_location: 'Union[Config.Object, str]', recursive_paths: List[str] = None, non_recursive_paths: List[str] = None,
               max_filesize: str = ClientWorkAttributeValues.CertificateDiscovery.MaxFilesize.less_than_10k,
               pkcs12_extensions: List[str] = ClientWorkAttributeValues.CertificateDiscovery.Extensions.default_pkcs12_extensions,
               pkcs7_extensions: List[str] = ClientWorkAttributeValues.CertificateDiscovery.Extensions.default_pkcs7_extensions,
               pem_extensions: List[str] = ClientWorkAttributeValues.CertificateDiscovery.Extensions.default_pem_extensions,
               ibmcms_extensions: List[str] = ClientWorkAttributeValues.CertificateDiscovery.Extensions.default_ibmcms_extensions,
               jks_jceks_extensions: List[str] = ClientWorkAttributeValues.CertificateDiscovery.Extensions.default_jks_jceks_extensions,
               iplanet_extensions: List[str] = ClientWorkAttributeValues.CertificateDiscovery.Extensions.default_iplanet_extensions,
               exclude_recursive_paths: List[str] = None, exclude_non_recursive_paths: List[str] = None,
               exclude_file_patterns: List[str] = None, scan_mounted_file_systems: bool = False,
               log_threshold: str = ClientWorkAttributeValues.CertificateDiscovery.LogThreshold.info,
               attributes: dict = None, get_if_already_exists: bool = True):
        """
        Args:
            name: The name of the client work.
            certificate_location: :ref:`config_object` or :ref:`dn` of the folder to place certificates.
            recursive_paths: A list of file paths to recursively search for new certificates
            non_recursive_paths: A list of file paths to search for new certificates
            max_filesize: A maximum file size (Ignores files larger than this size)
            pkcs12_extensions: A list of pkcs#12 extensions to match (defaults to .p12, .pfx)
            pkcs7_extensions: A list of pkcs#7 extensions to match (defaults to .p7b, .p7c, .p7)
            pem_extensions: A list of PEM extensions to match (defaults to .cer, .der, .crt, .pem)
            ibmcms_extensions: A list of IBM CMS extensions to match (defaults to .kdb)
            jks_jceks_extensions: A list of JKS/JCKES(java) extensions to match (defaults to .jck, .jceks, .jks, cacerts)
            iplanet_extensions: A list of iPlanet(Berkeley/NSS) extensions to match (defaults to .db)
            exclude_recursive_paths: A list of file paths to exclude (recursively) from discovery
            exclude_non_recursive_paths: A list of file paths to exclude from discovery
            exclude_file_patterns: A list of file patterns to exclude from discovery
            scan_mounted_file_systems: Scan file systems mounted via NFS/CIFS/NTFS junction points (defaults to False)
            log_threshold: set the logging level (defaults to INFO)
            attributes: Additional attributes to apply to the object.
            get_if_already_exists: If the objects already exists, just return it as is.

        Returns:
            :ref:`config_object` of the client work.
        """
        certificate_location_dn = self._get_dn(certificate_location)
        work_attributes = {
            ClientCertificateDiscoveryWorkAttributes.created_by                 : ClientWorkAttributeValues.CertificateDiscovery.CreatedBy.websdk,
            ClientCertificateDiscoveryWorkAttributes.certificate_location_dn    : certificate_location_dn,
            ClientCertificateDiscoveryWorkAttributes.interval                   : 0,
            ClientCertificateDiscoveryWorkAttributes.log_threshold              : log_threshold,
            ClientCertificateDiscoveryWorkAttributes.exclude_remote_mount_points: int(scan_mounted_file_systems),
            ClientCertificateDiscoveryWorkAttributes.max_filesize               : max_filesize
        }

        paths = []

        for path in (non_recursive_paths or []):
            paths.append(f'1,{path}')
        for path in (recursive_paths or []):
            paths.append(f'2,{path}')
        for path in (exclude_non_recursive_paths or []):
            paths.append(f'3,{path}')
        for path in (exclude_recursive_paths or []):
            paths.append(f'4,{path}')
        for path in (exclude_file_patterns or []):
            paths.append(f'6,{path}')

        if len(paths) > 0:
            work_attributes[ClientCertificateDiscoveryWorkAttributes.certificate_scanner_path] = paths

        extensions = []

        for ext in (pkcs12_extensions or []):
            extensions.append(f'5,{ext}')
        for ext in (pkcs7_extensions or []):
            extensions.append(f'7,{ext}')
        for ext in (pem_extensions or []):
            extensions.append(f'8,{ext}')
        for ext in (ibmcms_extensions or []):
            extensions.append(f'3,{ext}')
        for ext in (jks_jceks_extensions or []):
            extensions.append(f'4,{ext}')
        for ext in (iplanet_extensions or []):
            extensions.append(f'2,{ext}')

        if len(extensions) > 0:
            work_attributes[ClientCertificateDiscoveryWorkAttributes.certificate_scanner_map] = extensions

        if attributes:
            work_attributes.update(attributes)

        return self._config_create(
            name=name,
            parent_folder_dn=self._work_base_dn,
            config_class=ClientCertificateDiscoveryWorkAttributes.__config_class__,
            attributes=work_attributes,
            keep_list_values=True,
            get_if_already_exists=get_if_already_exists
        )


    def unschedule(self, work: 'Union[Config.Object, str]'):
        """
        Removes any scheduling for the client work, but does not delete the client work.

        Args:
            work: :ref:`config_object` or name of the client work.
        """
        work_dn = self._get_dn(work, parent_dn=self._work_base_dn)
        for attribute_name in {
            ClientCertificateDiscoveryWorkAttributes.start_time,
            ClientCertificateDiscoveryWorkAttributes.schedule_type,
            ClientCertificateDiscoveryWorkAttributes.interval,
            ClientCertificateDiscoveryWorkAttributes.days_of_week,
            ClientCertificateDiscoveryWorkAttributes.days_of_month
        }:
            self._api.websdk.Config.ClearAttribute.post(
                object_dn=work_dn,
                attribute_name=attribute_name
            ).assert_valid_response()


@feature('Certificate Enrollment Via EST Protocol')
class CertificateEnrollmentViaESTProtocol(_ClientWorkBase):
    def __init__(self, api):
        super().__init__(api=api)

    def create(self, name: str, certificate_container: 'Union[Config.Object, str]', naming_pattern: str, ca_template: 'Union[Config.Object, str]',
               contacts: 'List[Identity.Identity, str]', certificate_origin: str = None, certificate_description: str = None,
               validation_type: int = ClientWorkAttributeValues.CertificateEnrollmentViaESTProtocol.ValidationType.basic,
               revocation_status_check: int = ClientWorkAttributeValues.CertificateEnrollmentViaESTProtocol.RevocationStatusCheck.accept_when_unknown,
               authentication_credentials: 'Union[Config.Object, str]' = None, authenticate_only_by_password: bool = False,
               revoke_previous_version: bool = False,
               identity_verification: int = ClientWorkAttributeValues.CertificateEnrollmentViaESTProtocol.IdentityVerification.valid,
               trusted_certs_and_cas: 'List[Config.Object, str]' = None, get_if_already_exists: bool = False, attributes: dict = None):
        """
        Args:
            name: The name of the client work.
            certificate_container: :ref:`config_object` or :ref:`dn` of the folder to create certificates.
            naming_pattern: The object naming pattern (IE. $CSR.CN$)
            ca_template: :ref:`config_object` or DN of the Certificate Authority.
            contacts: List of :ref:`identity_object` or :ref:`prefixed_name` of the contacts.
            certificate_origin: Specify the certificate origin value
            certificate_description: Specify the certificate description value
            validation_type:
                * basic: (default) Checks Expiration, Revocation, and Chain of Trust
                * strict: Performs Basic Validation and checks Client Authentication Enhanced Key Usage
            revocation_status_check: Defaults to accept when unknown.
            authentication_credentials: :ref:`config_object` or :ref:`dn` of the credential to provide client password authentication.
            authenticate_only_by_password: Only accept requests that are authenticated by password
            revoke_previous_version: Revoke previous versions of the certificate (defaults to False)
            identity_verification: Proof of Possession
            trusted_certs_and_cas: A List of :ref:`config_object` or :ref:`dn` of the Certificate Authorities and Certificates to trust.
            attributes: Additional attributes to apply to the object.
            get_if_already_exists: If the objects already exists, just return it as is.

        Returns:
            :ref:`config_object` of the client work.
        """
        certificate_container_dn = self._get_dn(certificate_container)
        ca_template_dn = self._get_dn(ca_template)
        work_attributes = {
            NetworkDeviceCertificateWorkAttributes.created_by                                 : ClientWorkAttributeValues.CertificateEnrollmentViaESTProtocol.CreatedBy.websdk,
            NetworkDeviceCertificateWorkAttributes.certificate_container                      : certificate_container_dn,
            NetworkDeviceCertificateWorkAttributes.naming_pattern                             : naming_pattern,
            NetworkDeviceCertificateWorkAttributes.certificate_authority                      : ca_template_dn,
            NetworkDeviceCertificateWorkAttributes.contact                                    : [
                self._get_prefixed_universal(c) for c in contacts] if contacts else None,
            NetworkDeviceCertificateWorkAttributes.client_certificate_eku_checks_enabled      : validation_type,
            NetworkDeviceCertificateWorkAttributes.revocation_mode                            : revocation_status_check,
            NetworkDeviceCertificateWorkAttributes.fallback_to_http_auth                      : int(authenticate_only_by_password == False),
            NetworkDeviceCertificateWorkAttributes.revoke_existing_certificate_on_reenrollment: int(revoke_previous_version),
            NetworkDeviceCertificateWorkAttributes.pop_mode                                   : identity_verification
        }

        if certificate_origin: work_attributes[NetworkDeviceCertificateWorkAttributes.origin] = certificate_origin
        if certificate_description: work_attributes[NetworkDeviceCertificateWorkAttributes.description] = certificate_description
        if authentication_credentials: 
            work_attributes[NetworkDeviceCertificateWorkAttributes.authentication_credentials] = self._get_dn(authentication_credentials)

        if trusted_certs_and_cas:
            work_attributes[NetworkDeviceCertificateWorkAttributes.explicit_trust_anchors] = [
                self._get_dn(c) for c in trusted_certs_and_cas
            ]
            work_attributes[NetworkDeviceCertificateWorkAttributes.use_implicit_trust_anchors] = 0
        else:
            work_attributes[NetworkDeviceCertificateWorkAttributes.use_implicit_trust_anchors] = 1

        if attributes:
            work_attributes.update(attributes)

        return self._config_create(
            name=name,
            parent_folder_dn=self._work_base_dn,
            config_class=NetworkDeviceCertificateWorkAttributes.__config_class__,
            attributes=work_attributes,
            keep_list_values=True,
            get_if_already_exists=get_if_already_exists
        )


@feature('Certificate Installation')
class CertificateInstallation(_ClientWorkBase):
    def __init__(self, api):
        super().__init__(api=api)

    def schedule(self, work: 'Union[Config.Object, str]', start_time: int = None, daily: bool = False, hourly: bool = False,
                 on_receipt: bool = False,
                 days_of_week: List[str] = None,
                 days_of_month: List[str] = None, every_x_minutes: int = None, randomize_minutes: int = 0):
        """
        .. note::
			Only one of daily, hourly, on_receipt, days_of_week, days_of_month or every_x_minutes can be set.
			
        Schedules the Certificate Installation work to run.

        Args:
            work: :ref:`config_object` or name of the client work.
            start_time: The 24-hour UTC hour format (i.e. 20 = 8PM UTC) for the job to start.
            daily: Runs the client work daily
            hourly: Runs the client work hourly
            on_receipt: Runs the client work on receipt
            days_of_week: Runs the client work on specific days of the week. It is a Zero-based index of the days of the week (i.e. Sunday = '0').
            days_of_month: Runs the client work on specific days of the month.
            every_x_minutes: Runs the client work every 1,5,15 or 30 minutes. (Must be one of 1, 5, 15 or 30)
            randomize_minutes: Randomize the given minutes for agent to send data back to the server
        """
        work_dn = self._get_dn(work, parent_dn=self._work_base_dn)
        attributes = {
            ClientCertificateDiscoveryWorkAttributes.interval: randomize_minutes
        }

        if len([x for x in [daily, hourly, on_receipt, days_of_week, days_of_month, every_x_minutes] if
                x not in [None, False]]) != 1:
            raise InvalidFormat(
                "Error in Schedule: must specify one (and only one) of: daily,hourly,on_receipt,days_of_week,days_of_month,every_x_minutes")

        if daily:
            if not start_time:
                raise InvalidFormat(
                    "Error in Schedule: must supply a 'start_time' to use the daily interval")

            attributes[ClientCertificateDiscoveryWorkAttributes.start_time] = datetime.time(
                start_time % 24).strftime(
                "%I:00 %p")
            attributes[ClientCertificateDiscoveryWorkAttributes.schedule_type] = \
                ClientWorkAttributeValues.CertificateInstallation.ScheduleType.daily
        elif hourly:
            attributes[ClientCertificateDiscoveryWorkAttributes.schedule_type] = \
                ClientWorkAttributeValues.CertificateInstallation.ScheduleType.hourly
        elif days_of_week:
            if not start_time:
                raise InvalidFormat(
                    "Error in Schedule: must supply a 'start_time' to use the days_of_week interval")
            attributes[ClientCertificateDiscoveryWorkAttributes.schedule_type] = \
                ClientWorkAttributeValues.CertificateInstallation.ScheduleType.days_of_week
            attributes[ClientCertificateDiscoveryWorkAttributes.days_of_week] = days_of_week
        elif days_of_month:
            if not start_time:
                raise InvalidFormat(
                    "Error in Schedule: must supply a 'start_time' to use the days_of_month interval")
            attributes[ClientCertificateDiscoveryWorkAttributes.schedule_type] = ClientWorkAttributeValues.CertificateInstallation.ScheduleType.days_of_month
            attributes[ClientCertificateDiscoveryWorkAttributes.days_of_month] = days_of_month
        elif on_receipt:
            attributes[ClientCertificateDiscoveryWorkAttributes.schedule_type] = ClientWorkAttributeValues.CertificateInstallation.ScheduleType.on_receipt
        elif every_x_minutes:
            if every_x_minutes == 30:
                attributes[ClientCertificateDiscoveryWorkAttributes.schedule_type] = ClientWorkAttributeValues.CertificateInstallation.ScheduleType.every_x_minutes
                attributes[ClientCertificateDiscoveryWorkAttributes.start_time] = "12:30:00 AM"
            elif every_x_minutes == 15:
                attributes[ClientCertificateDiscoveryWorkAttributes.schedule_type] = ClientWorkAttributeValues.CertificateInstallation.ScheduleType.every_x_minutes
                attributes[ClientCertificateDiscoveryWorkAttributes.start_time] = "12:15:00 AM"
            elif every_x_minutes == 5:
                attributes[ClientCertificateDiscoveryWorkAttributes.schedule_type] = ClientWorkAttributeValues.CertificateInstallation.ScheduleType.every_x_minutes
                attributes[ClientCertificateDiscoveryWorkAttributes.start_time] = "12:05:00 AM"
            elif every_x_minutes == 1:
                attributes[ClientCertificateDiscoveryWorkAttributes.schedule_type] = ClientWorkAttributeValues.CertificateInstallation.ScheduleType.every_x_minutes
                attributes[ClientCertificateDiscoveryWorkAttributes.start_time] = "12:01:00 AM"
            else:
                raise InvalidFormat(
                    "Error in Schedule: must supply at one of (30, 15, 5, 1) for every_x_minutes")
        else:
            raise InvalidFormat(
                "Error in Schedule: must supply at one of (daily,hourly,on_receipt,days_of_week,days_of_month,every_x_minutes)")

        response = self._api.websdk.Config.Write.post(
            object_dn=work_dn,
            attribute_data=self._name_value_list(attributes)
        )
        response.assert_valid_response()

    def create(self, name: str, log_threshold: str = ClientWorkAttributeValues.CertificateInstallation.LogThreshold.info,
               attributes: dict = None, get_if_already_exists: bool = True):
        """
        Args:
            name: The name of the client work.
            log_threshold: set the logging level (defaults to INFO)
            attributes: Additional attributes to apply to the object.
            get_if_already_exists: If the objects already exists, just return it as is.

        Returns:
            :ref:`config_object` of the client work.
        """
        work_attributes = {
            CertificateProvisioningWorkAttributes.created_by   : ClientWorkAttributeValues.CertificateInstallation.CreatedBy.websdk,
            CertificateProvisioningWorkAttributes.interval     : 0,
            CertificateProvisioningWorkAttributes.log_threshold: log_threshold
        }

        if attributes:
            work_attributes.update(attributes)

        return self._config_create(
            name=name,
            parent_folder_dn=self._work_base_dn,
            config_class=CertificateProvisioningWorkAttributes.__config_class__,
            attributes=work_attributes,
            get_if_already_exists=get_if_already_exists
        )

    def unschedule(self, work: 'Union[Config.Object, str]'):
        """
        Removes any scheduling for the client work, but does not delete the client work.

        Args:
            work: :ref:`config_object` or name of the client work.
        """
        work_dn = self._get_dn(work, parent_dn=self._work_base_dn)
        for attribute_name in {
            ClientCertificateDiscoveryWorkAttributes.start_time,
            ClientCertificateDiscoveryWorkAttributes.schedule_type,
            ClientCertificateDiscoveryWorkAttributes.interval,
            ClientCertificateDiscoveryWorkAttributes.days_of_week,
            ClientCertificateDiscoveryWorkAttributes.days_of_month
        }:
            self._api.websdk.Config.ClearAttribute.post(
                object_dn=work_dn,
                attribute_name=attribute_name
            ).assert_valid_response()


@feature('Device Certificate Creation')
class DeviceCertificateCreation(_ClientWorkBase):
    def __init__(self, api):
        super().__init__(api=api)

    def create(self, name: str, certificate_container: 'Union[Config.Object, str]', ca_template: 'Union[Config.Object, str]',
               contacts: 'List[Identity.Identity, str]', description: str = None, naming_pattern: str = "$Client.DNSName$",
               common_name: str = "$Client.DNSName$", organization: str = None,
               organizational_unit: List[str] = None, city_locality: str = None, state_province: str = None,
               country: str = None, subject_alternative_names: bool = False, automatic_renewal: bool = True,
               renewal_days_before: int = 30, key_bit_strength: int = 2048, allow_certificate_sharing: bool = False,
               attributes: dict = None, get_if_already_exists: bool = True):
        """
        Args:
            name: The name of the client work
            certificate_container: :ref:`config_object` or :ref:`dn` of the folder to place certificates.
            ca_template: :ref:`config_object` or :ref:`dn` of the Certificate Authority.
            contacts: List of :ref:`identity_object` or :ref:`prefixed_name` of the contacts.
            description: A description for the certificates
            naming_pattern: Object naming pattern (defaults to: $Client.DNSName$)
            common_name: Common name for the certificate
            organization: Organization for the certificate
            organizational_unit: A list of organizational units for the certificate
            city_locality: A city or locality for the certificate
            state_province: A state or province for the certificate
            country: A country code for the certificate
            subject_alternative_names: A subject alternative name for the certificate
            automatic_renewal: Enable automatic renewal (defaults to True)
            renewal_days_before: Days before expiration for automatic renewal (defaults to 30)
            key_bit_strength: Key size of the certificates (defaults to 2048)
            allow_certificate_sharing: Allow sharing with mobile devices (default sto False)
            attributes: Additional attributes to apply to the object.
            get_if_already_exists: If the objects already exists, just return it as is.

        Returns:
            :ref:`config_object` of the client work.
        """
        certificate_container_dn = self._get_dn(certificate_container)
        ca_template_dn = self._get_dn(ca_template)
        
        work_attributes = {
            ClientUserCertificateWorkAttributes.created_by               : ClientWorkAttributeValues.DeviceCertificateCreation.CreatedBy.websdk,
            ClientUserCertificateWorkAttributes.certificate_container    : certificate_container_dn,
            ClientUserCertificateWorkAttributes.certificate_authority    : ca_template_dn,
            ClientUserCertificateWorkAttributes.naming_pattern           : naming_pattern,
            ClientUserCertificateWorkAttributes.contact                  : [self._get_prefixed_universal(c) for c in
                                                                            contacts] if contacts else None,
            ClientUserCertificateWorkAttributes.key_bit_strength         : key_bit_strength,
            ClientUserCertificateWorkAttributes.x509_subject             : common_name,
            ClientUserCertificateWorkAttributes.disable_automatic_renewal: not automatic_renewal,
            ClientUserCertificateWorkAttributes.transfer_allowed         : allow_certificate_sharing
        }

        if attributes:
            work_attributes.update(attributes)

        if description: work_attributes[ClientCertificateWorkAttributes.description] = description
        if organization: work_attributes[ClientCertificateWorkAttributes.organization] = organization
        if city_locality: work_attributes[ClientCertificateWorkAttributes.city] = city_locality
        if state_province: work_attributes[ClientCertificateWorkAttributes.state] = state_province
        if organizational_unit: work_attributes[ClientCertificateWorkAttributes.organizational_unit] = organizational_unit
        if country: work_attributes[ClientCertificateWorkAttributes.country] = country
        if subject_alternative_names: work_attributes[ClientCertificateWorkAttributes.x509_subjectaltname_dns] = common_name
        if automatic_renewal: work_attributes[ClientCertificateWorkAttributes.renewal_window] = renewal_days_before

        return self._config_create(
            name=name,
            parent_folder_dn=self._work_base_dn,
            config_class=ClientCertificateWorkAttributes.__config_class__,
            attributes=work_attributes,
            keep_list_values=True,
            get_if_already_exists=get_if_already_exists
        )


@feature('Dynamic Provisioning')
class DynamicProvisioning(_ClientWorkBase):
    def __init__(self, api):
        super().__init__(api=api)

    def create(self, name: str, certificate_container: 'Union[Config.Object, str]', ca_template: 'Union[Config.Object, str]',
               contacts: 'List[Identity.Identity, str]', description: str = None, naming_pattern: str = "$Client.DNSName$", common_name: str = "$Client.DNSName$",
               organization: str = None, organizational_unit: List[str] = None, city_locality: str = None,
               state_province: str = None, country: str = None, subject_alternative_names: str = "$Client.DNSname$",
               capi_keystore: bool = False, capi_friendly_name: str = "", capi_trustee: str = "",
               key_bit_strength: int = 2048, retry_interval: int = 15,
               log_threshold: str = ClientWorkAttributeValues.DynamicProvisioning.LogThreshold.info,
               attributes: dict = None, get_if_already_exists: bool = True):
        """
        Args:
            name: The name of the client work
            certificate_container: :ref:`config_object` or :ref:`dn` of the folder to place certificates.
            ca_template: :ref:`config_object` or DN of the Certificate Authority.
            contacts: List of :ref:`identity_object` or :ref:`prefixed_name` of the contacts.
            description: A description for the certificates
            naming_pattern: object naming pattern (defaults to: $Client.DNSName$)
            common_name: Common name for the certificate
            organization: Organization for the certificate
            organizational_unit: A list of organizational units for the certificate
            city_locality: A city or locality for the certificate
            state_province: A state or province for the certificate
            country: A country code for the certificate
            subject_alternative_names: A subject alternative name for the certificate
            capi_keystore: Use a capi keystore (defaults to False)
            capi_friendly_name: Friendly name in the keystore
            capi_trustee: Set the capi trustee
            key_bit_strength: Key size for the certificate (defaults to 2048)
            retry_interval: An interval in minutes (15, 30, 45, 60) for the agent to retry
            log_threshold: Set the logging level (defaults to INFO)
            attributes: Additional attributes to apply to the object.
            get_if_already_exists: If the objects already exists, just return it as is.

        Returns:
            :ref:`config_object` of the client work.
        """
        certificate_container_dn = self._get_dn(certificate_container)
        ca_template_dn = self._get_dn(ca_template)
        
        work_attributes = {
            ServerCertificateWorkAttributes.created_by             : ClientWorkAttributeValues.DynamicProvisioning.CreatedBy.websdk,
            ServerCertificateWorkAttributes.certificate_container  : certificate_container_dn,
            ServerCertificateWorkAttributes.certificate_authority  : ca_template_dn,
            ServerCertificateWorkAttributes.naming_pattern         : naming_pattern,
            ServerCertificateWorkAttributes.contact                : [self._get_prefixed_universal(c) for c in contacts] if contacts else None,
            ServerCertificateWorkAttributes.key_bit_strength       : key_bit_strength,
            ServerCertificateWorkAttributes.x509_subject           : common_name,
            ServerCertificateWorkAttributes.x509_subjectaltname_dns: subject_alternative_names,
            ServerCertificateWorkAttributes.interval               : retry_interval,
            ServerCertificateWorkAttributes.log_threshold          : log_threshold
        }

        if description: work_attributes[ServerCertificateWorkAttributes.description] = description
        if organization: work_attributes[ServerCertificateWorkAttributes.organization] = organization
        if city_locality: work_attributes[ServerCertificateWorkAttributes.city] = city_locality
        if state_province: work_attributes[ServerCertificateWorkAttributes.state] = state_province
        if organizational_unit: work_attributes[ServerCertificateWorkAttributes.organizational_unit] = organizational_unit
        if country: work_attributes[ServerCertificateWorkAttributes.country] = country
        if capi_keystore:
            work_attributes[ServerCertificateWorkAttributes.application_type] = ClientWorkAttributeValues.DynamicProvisioning.ApplicationType.capi
            work_attributes[ServerCertificateWorkAttributes.friendly_name] = capi_friendly_name
            work_attributes[ServerCertificateWorkAttributes.private_key_trustee] = capi_trustee

        if attributes:
            work_attributes.update(attributes)

        return self._config_create(
            name=name,
            parent_folder_dn=self._work_base_dn,
            config_class=ServerCertificateWorkAttributes.__config_class__,
            attributes=work_attributes,
            keep_list_values=True,
            get_if_already_exists=get_if_already_exists
        )


@feature('SSH Device Placement')
class SSHDevicePlacement(_ClientWorkBase):
    def __init__(self, api):
        super().__init__(api=api)

    def create(self, name: str, devices_folder: 'Union[Config.Object, str]',
               share_mode: str = ClientWorkAttributeValues.SSHDevicePlacement.DeviceSharedMode.devices_folder_and_sub_folders,
               attributes: dict = None, get_if_already_exists: bool = True):
        """
        Args:
            name: The name of the client work
            devices_folder: :ref:`config_object` or :ref:`dn` of the folder to place newly discovered ssh devices.
            share_mode: Mode to de-duplicate newly discovered devices:
                * "WholeTree" : Search the entire policy tree
                * "SpecifiedFolderOnly" : Search the devices folder
                * "SpecifiedFolderAndSubFolders" : Search the devices folder and all sub-folders
                * "None" : Create a duplicate device
            attributes: Additional attributes to apply to the object.
            get_if_already_exists: If the objects already exists, just return it as is.

        Returns:
            :ref:`config_object` of the client work.
        """
        devices_folder_dn = self._get_dn(devices_folder)
        work_attributes = {
            ServerAgentSSHDevicePlacementWorkAttributes.created_by            : ClientWorkAttributeValues.AgentConnectivity.CreatedBy.websdk,
            ServerAgentSSHDevicePlacementWorkAttributes.device_object_location: devices_folder_dn,
            ServerAgentSSHDevicePlacementWorkAttributes.device_share_mode     : share_mode
        }

        if attributes:
            work_attributes.update(attributes)

        return self._config_create(
            name=name,
            parent_folder_dn=self._work_base_dn,
            config_class=ServerAgentSSHDevicePlacementWorkAttributes.__config_class__,
            attributes=work_attributes,
            get_if_already_exists=get_if_already_exists
        )


@feature('SSH Discovery')
class SSHDiscovery(_ClientWorkBase):
    def __init__(self, api):
        super().__init__(api=api)

    def schedule(self, work: 'Union[Config.Object, str]', start_time: int = None, daily: bool = False, hourly: bool = False,
                 on_receipt: bool = False, every_30_minutes: bool = False,
                 days_of_week: List[str] = None,
                 days_of_month: List[str] = None, randomize_minutes: int = 0, full_scan: bool = False):
        """
        .. note::
			Only one of daily, hourly, on_receipt, every_30_minutes, days_of_week or days_of_month can be set.

        Schedules the SSH Discovery work to run.

        Args:
            work: :ref:`config_object` or name of the client work.
            start_time: The 24-hour UTC hour format (i.e. 20 = 8PM UTC) for the job to start.
            daily: Runs the client work daily
            hourly: Runs the client work hourly
            on_receipt: Runs the client work on_receipt
            every_30_minutes: Runs the client work every 30 minutes
            days_of_week: Runs the client work on specific days of the week. It is a Zero-based index of the days of the week (i.e. Sunday = '0').
            days_of_month: Runs the client work on specific days of the month.
            randomize_minutes: Randomize the given minutes for agent to send data back to the server
            full_scan: Reset the cache and perform a full scan (resend all the data to the server)
        """
        work_dn = self._get_dn(work, parent_dn=self._work_base_dn)
        attributes = {
            ClientAgentSSHDiscoveryWorkAttributes.interval: randomize_minutes
        }

        if len([x for x in [daily, hourly, on_receipt, every_30_minutes, days_of_week, days_of_month] if x not in [None, False]]) != 1:
            raise InvalidFormat(
                "Error in Schedule: must specify one (and only one) of: daily,hourly,on_receipt,every_30_minutes,days_of_week,days_of_month")

        if full_scan:
            attributes[ClientAgentSSHDiscoveryWorkAttributes.clear_cache_timestamp] = str(time.time())

        if daily:
            if not start_time:
                raise InvalidFormat(
                    "Error in Schedule: must supply a 'start_time' to use the daily interval")

            attributes[ClientAgentSSHDiscoveryWorkAttributes.start_time] = \
                datetime.time(start_time % 24).strftime("%I:00 %p")
            attributes[ClientAgentSSHDiscoveryWorkAttributes.schedule_type] = \
                ClientWorkAttributeValues.SSHDiscovery.ScheduleType.daily
        elif hourly:
            attributes[ClientAgentSSHDiscoveryWorkAttributes.schedule_type] = \
                ClientWorkAttributeValues.SSHDiscovery.ScheduleType.hourly
        elif on_receipt:
            attributes[ClientAgentSSHDiscoveryWorkAttributes.schedule_type] = \
                ClientWorkAttributeValues.SSHDiscovery.ScheduleType.on_receipt
        elif every_30_minutes:
            attributes[ClientAgentSSHDiscoveryWorkAttributes.schedule_type] = \
                ClientWorkAttributeValues.SSHDiscovery.ScheduleType.every_30_minutes
            attributes[ClientAgentSSHDiscoveryWorkAttributes.start_time] = "12:30:00 AM"
        elif days_of_week:
            if not start_time:
                raise InvalidFormat(
                    "Error in Schedule: must supply a 'start_time' to use the days_of_week interval")

            attributes[ClientAgentSSHDiscoveryWorkAttributes.schedule_type] = \
                ClientWorkAttributeValues.SSHDiscovery.ScheduleType.days_of_week
            attributes[ClientAgentSSHDiscoveryWorkAttributes.days_of_week] = days_of_week
        elif days_of_month:
            if not start_time:
                raise InvalidFormat(
                    "Error in Schedule: must supply a 'start_time' to use the days_of_month interval")

            attributes[ClientAgentSSHDiscoveryWorkAttributes.schedule_type] = \
                ClientWorkAttributeValues.SSHDiscovery.ScheduleType.days_of_month
            attributes[ClientAgentSSHDiscoveryWorkAttributes.days_of_month] = days_of_month
        else:
            raise InvalidFormat(
                "Error in Schedule: must supply one of (daily,hourly,on_receipt,every_30_minutes,days_of_week,days_of_month)")

        response = self._api.websdk.Config.Write.post(
            object_dn=work_dn,
            attribute_data=self._name_value_list(attributes, True)
        )
        response.assert_valid_response()

    def create(self, name: str, scan_default_paths: bool = True, host_key_paths: List[str] = None,
               user_key_paths: List[str] = None,
               user_or_host_paths: List[str] = None, exclude_paths: List[str] = None, scan_mounted_fs: bool = False,
               minimize_resources: bool = False,
               max_filesize: int = ClientWorkAttributeValues.SSHDiscovery.MaxFilesize.less_than_1MB,
               log_threshold: str = ClientWorkAttributeValues.SSHDiscovery.LogThreshold.info,
               attributes: dict = None, get_if_already_exists: bool = True):
        """
        Args:
            name: The name of the client work
            scan_default_paths: scan all of the default paths for ssh keys
            host_key_paths: A list of paths to scan for host keys
            user_key_paths: A list of paths to scan for user keys
            user_or_host_paths: A list of paths to scan for both host and user keys
            exclude_paths: A list of paths to exclude from scan
            scan_mounted_fs: Scan file systems mounted via NFS/CIFS/NTFS junction points (defaults to False)
            minimize_resources: Minimizes resource usage during scan (defaults to False)
            max_filesize: Ignore files larger than this size (defaults to 1MB)
            log_threshold: Set the logging level (defaults to INFO)
            attributes: Additional attributes to apply to the object.
            get_if_already_exists: If the objects already exists, just return it as is.

        Returns:
            :ref:`config_object` of the client work.
        """
        work_attributes = {
            ClientAgentSSHDiscoveryWorkAttributes.created_by                   : ClientWorkAttributeValues.SSHDiscovery.CreatedBy.websdk,
            ClientAgentSSHDiscoveryWorkAttributes.interval                     : 0,
            ClientAgentSSHDiscoveryWorkAttributes.server_path_defaults_disabled: int(not scan_default_paths),
            ClientAgentSSHDiscoveryWorkAttributes.user_path_defaults_disabled  : int(not scan_default_paths),
            ClientAgentSSHDiscoveryWorkAttributes.exclude_remote_mount_points  : int(not scan_mounted_fs),
            ClientAgentSSHDiscoveryWorkAttributes.minimize_resource_use        : int(minimize_resources),
            ClientAgentSSHDiscoveryWorkAttributes.max_filesize                 : max_filesize,
            ClientAgentSSHDiscoveryWorkAttributes.log_threshold                : log_threshold
        }

        scanner_paths = []
        user_paths = []

        for path in (host_key_paths or []):
            scanner_paths.append(f'2,{path}')
        for path in (user_key_paths or []):
            user_paths.append(f'2,{path}')
        for path in (user_or_host_paths or []):
            scanner_paths.append(f'2,{path}')
            user_paths.append(f'2,{path}')
        for path in (exclude_paths or []):
            scanner_paths.append(f'4,{path}')
            user_paths.append(f'4,{path}')

        if len(scanner_paths) > 0: work_attributes[ClientAgentSSHDiscoveryWorkAttributes.ssh_scanner_service_path]: scanner_paths
        if len(user_paths) > 0: work_attributes[ClientAgentSSHDiscoveryWorkAttributes.ssh_scanner_user_path]: user_paths

        if attributes:
            work_attributes.update(attributes)

        return self._config_create(
            name=name,
            parent_folder_dn=self._work_base_dn,
            config_class=ClientAgentSSHDiscoveryWorkAttributes.__config_class__,
            attributes=work_attributes,
            keep_list_values=True,
            get_if_already_exists=get_if_already_exists
        )

    def unschedule(self, work: 'Union[Config.Object, str]'):
        """
        Removes any scheduling for the client work, but does not delete the client work.

        Args:
            work: :ref:`config_object` or name of the client work.
        """
        work_dn = self._get_dn(work, parent_dn=self._work_base_dn)
        for attribute_name in {
            ClientAgentSSHDiscoveryWorkAttributes.start_time,
            ClientAgentSSHDiscoveryWorkAttributes.schedule_type,
            ClientAgentSSHDiscoveryWorkAttributes.interval,
            ClientAgentSSHDiscoveryWorkAttributes.days_of_week,
            ClientAgentSSHDiscoveryWorkAttributes.days_of_month
        }:
            self._api.websdk.Config.ClearAttribute.post(
                object_dn=work_dn,
                attribute_name=attribute_name
            ).assert_valid_response()


@feature('SSH Key Usage')
class SSHKeyUsage(_ClientWorkBase):
    def __init__(self, api):
        super().__init__(api=api)

    def schedule(self, work: 'Union[Config.Object, str]', start_time: int = None, daily: bool = False, hourly: bool = False,
                 on_receipt: bool = False, every_x_minutes: int = None, randomize_minutes: int = 0):
        """
        .. note::
			Only one of daily, hourly, on_receipt, or every_x_minutes can be set.

        Schedules the SSH KeyUsage work to run.

        Args:
            work: :ref:`config_object` or name of the client work.
            start_time: The 24-hour UTC hour format (i.e. 20 = 8PM UTC) for the job to start.
            daily: Runs the client work daily
            hourly: Runs the client work hourly
            on_receipt: Runs the client work on_receipt
            every_x_minutes: Runs the client work every: 1, 5, 15 or 30 minutes
            randomize_minutes: Randomize the given minutes for agent to send data back to the server
        """
        work_dn = self._get_dn(work, parent_dn=self._work_base_dn)
        attributes = {
            ClientAgentSSHKeyUsageWorkAttributes.interval: randomize_minutes
        }

        if len([x for x in [daily, hourly, on_receipt, every_x_minutes] if x not in [None, False]]) != 1:
            raise InvalidFormat(
                "Error in Schedule: must specify one (and only one) of: daily,hourly,on_receipt, every_x_minutes")

        if daily:
            if not start_time:
                raise InvalidFormat(
                    "Error in Schedule: must supply a 'start_time' to use the daily interval")

            attributes[ClientAgentSSHKeyUsageWorkAttributes.start_time] =\
                datetime.time(start_time % 24).strftime("%I:00 %p")
            attributes[ClientAgentSSHKeyUsageWorkAttributes.schedule_type] = \
                ClientWorkAttributeValues.SSHKeyUsage.ScheduleType.daily
        elif hourly:
            attributes[ClientAgentSSHKeyUsageWorkAttributes.schedule_type] = \
                ClientWorkAttributeValues.SSHKeyUsage.ScheduleType.hourly
        elif on_receipt:
            attributes[ClientAgentSSHKeyUsageWorkAttributes.schedule_type] = \
                ClientWorkAttributeValues.SSHKeyUsage.ScheduleType.on_receipt
        elif every_x_minutes:
            if every_x_minutes == 30:
                attributes[ClientAgentSSHKeyUsageWorkAttributes.schedule_type] = \
                    ClientWorkAttributeValues.SSHKeyUsage.ScheduleType.every_x_minutes
                attributes[ClientAgentSSHKeyUsageWorkAttributes.start_time] = "12:30:00 AM"
            elif every_x_minutes == 15:
                attributes[ClientAgentSSHKeyUsageWorkAttributes.schedule_type] = \
                    ClientWorkAttributeValues.SSHKeyUsage.ScheduleType.every_x_minutes
                attributes[ClientAgentSSHKeyUsageWorkAttributes.start_time] = "12:15:00 AM"
            elif every_x_minutes == 5:
                attributes[ClientAgentSSHKeyUsageWorkAttributes.schedule_type] = \
                    ClientWorkAttributeValues.SSHKeyUsage.ScheduleType.every_x_minutes
                attributes[ClientAgentSSHKeyUsageWorkAttributes.start_time] = "12:05:00 AM"
            elif every_x_minutes == 1:
                attributes[ClientAgentSSHKeyUsageWorkAttributes.schedule_type] = \
                    ClientWorkAttributeValues.SSHKeyUsage.ScheduleType.every_x_minutes
                attributes[ClientAgentSSHKeyUsageWorkAttributes.start_time] = "12:01:00 AM"
            else:
                raise InvalidFormat(
                    "Error in Schedule: must supply at one of (30, 15, 5, 1) for every_x_minutes")
        else:
            raise InvalidFormat(
                "Error in Schedule: must supply one of (daily, hourly, on_receipt, every_x_minutes)")

        response = self._api.websdk.Config.Write.post(
            object_dn=work_dn,
            attribute_data=self._name_value_list(attributes)
        )
        response.assert_valid_response()

    def create(self, name: str, limit_cache_size: int = 50000,
               log_threshold: str = ClientWorkAttributeValues.SSHKeyUsage.LogThreshold.info,
               attributes: dict = None, get_if_already_exists: bool = True):
        """
        Args:
            name: The name of the client work.
            limit_cache_size: maximum items in the cache
            log_threshold: Set the logging level (defaults to INFO)
            attributes: Additional attributes to apply to the object.
            get_if_already_exists: If the objects already exists, just return it as is.

        Returns:
            :ref:`config_object` of the client work.
        """
        work_attributes = {
            ClientAgentSSHKeyUsageWorkAttributes.created_by   : ClientWorkAttributeValues.SSHKeyUsage.CreatedBy.websdk,
            ClientAgentSSHKeyUsageWorkAttributes.interval     : 0,
            ClientAgentSSHKeyUsageWorkAttributes.log_threshold: log_threshold,
            ClientAgentSSHKeyUsageWorkAttributes.max_row_count: limit_cache_size
        }

        if attributes:
            work_attributes.update(attributes)

        return self._config_create(
            name=name,
            parent_folder_dn=self._work_base_dn,
            config_class=ClientAgentSSHKeyUsageWorkAttributes.__config_class__,
            attributes=work_attributes,
            get_if_already_exists=get_if_already_exists
        )

    def unschedule(self, work: 'Union[Config.Object, str]'):
        """
        Removes any scheduling for the client work, but does not delete the client work.

        Args:
            work: :ref:`config_object` or name of the client work.
        """
        work_dn = self._get_dn(work, parent_dn=self._work_base_dn)
        for attribute_name in {
            ClientAgentSSHKeyUsageWorkAttributes.start_time,
            ClientAgentSSHKeyUsageWorkAttributes.schedule_type,
            ClientAgentSSHKeyUsageWorkAttributes.interval,
            ClientAgentSSHKeyUsageWorkAttributes.days_of_week,
            ClientAgentSSHKeyUsageWorkAttributes.days_of_month
        }:
            self._api.websdk.Config.ClearAttribute.post(
                object_dn=work_dn,
                attribute_name=attribute_name
            ).assert_valid_response()


@feature('SSH Remediation')
class SSHRemediation(_ClientWorkBase):
    def __init__(self, api):
        super().__init__(api=api)

    def schedule(self, work: 'Union[Config.Object, str]', start_time: int = None, daily: bool = False, hourly: bool = False,
                 on_receipt: bool = False, days_of_week: List[str] = None, days_of_month: List[str] = None,
                 every_x_minutes: int = None, randomize_minutes: int = 0):
        """
        .. note::
			Only one of daily, hourly, on_receipt, days_of_week, days_of_month or every_x_minutes can be set.

        Schedules the SSH Remediation work to run.

        Args:
            work: :ref:`config_object` or name of the client work.
            start_time: The 24-hour UTC hour format (i.e. 20 = 8PM UTC) for the job to start.
            daily: Runs the client work daily
            hourly: Runs the client work hourly
            on_receipt: Runs the client work on_receipt
            days_of_week: Runs the client work on specific days of the week. It is a Zero-based index of the days of the week (i.e. Sunday = '0').
            days_of_month: Runs the client work on specific days of the month.
            every_x_minutes: Runs the client work every: 1, 5, 15 or 30 minutes
            randomize_minutes: Randomize the given minutes for agent to send data back to the server
        """
        work_dn = self._get_dn(work, parent_dn=self._work_base_dn)
        attributes = {
            ClientAgentSSHDiscoveryWorkAttributes.interval: randomize_minutes
        }

        if len([x for x in [daily, hourly, on_receipt, days_of_week, days_of_month, every_x_minutes] if x not in [None, False]]) != 1:
            raise InvalidFormat(
                "Error in Schedule: must specify one (and only one) of: daily,hourly,on_receipt,days_of_week,days_of_month,every_x_minutes")

        if daily:
            if not start_time:
                raise InvalidFormat(
                    "Error in Schedule: must supply a 'start_time' to use the daily interval")

            attributes[ClientAgentSSHDiscoveryWorkAttributes.start_time] = \
                datetime.time(start_time % 24).strftime("%I:00 %p")
            attributes[ClientAgentSSHDiscoveryWorkAttributes.schedule_type] = \
                ClientWorkAttributeValues.SSHRemediation.ScheduleType.daily
        elif hourly:
            attributes[ClientAgentSSHDiscoveryWorkAttributes.schedule_type] = \
                ClientWorkAttributeValues.SSHRemediation.ScheduleType.hourly
        elif days_of_week:
            if not start_time:
                raise InvalidFormat(
                    "Error in Schedule: must supply a 'start_time' to use the days_of_week interval")
            attributes[ClientAgentSSHDiscoveryWorkAttributes.schedule_type] = \
                ClientWorkAttributeValues.SSHRemediation.ScheduleType.days_of_week
            attributes[ClientAgentSSHDiscoveryWorkAttributes.days_of_week] = days_of_week
        elif days_of_month:
            if not start_time:
                raise InvalidFormat(
                    "Error in Schedule: must supply a 'start_time' to use the days_of_month interval")
            attributes[ClientAgentSSHDiscoveryWorkAttributes.schedule_type] = \
                ClientWorkAttributeValues.SSHRemediation.ScheduleType.days_of_month
            attributes[ClientAgentSSHDiscoveryWorkAttributes.days_of_month] = days_of_month
        elif on_receipt:
            attributes[ClientAgentSSHDiscoveryWorkAttributes.schedule_type] = \
                ClientWorkAttributeValues.SSHRemediation.ScheduleType.on_receipt
        elif every_x_minutes:
            if every_x_minutes == 30:
                attributes[ClientAgentSSHDiscoveryWorkAttributes.schedule_type] = \
                    ClientWorkAttributeValues.SSHRemediation.ScheduleType.every_x_minutes
                attributes[ClientAgentSSHDiscoveryWorkAttributes.start_time] = "12:30:00 AM"
            elif every_x_minutes == 15:
                attributes[ClientAgentSSHDiscoveryWorkAttributes.schedule_type] = \
                    ClientWorkAttributeValues.SSHRemediation.ScheduleType.every_x_minutes
                attributes[ClientAgentSSHDiscoveryWorkAttributes.start_time] = "12:15:00 AM"
            elif every_x_minutes == 5:
                attributes[ClientAgentSSHDiscoveryWorkAttributes.schedule_type] = \
                    ClientWorkAttributeValues.SSHRemediation.ScheduleType.every_x_minutes
                attributes[ClientAgentSSHDiscoveryWorkAttributes.start_time] = "12:05:00 AM"
            elif every_x_minutes == 1:
                attributes[ClientAgentSSHDiscoveryWorkAttributes.schedule_type] = \
                    ClientWorkAttributeValues.SSHRemediation.ScheduleType.every_x_minutes
                attributes[ClientAgentSSHDiscoveryWorkAttributes.start_time] = "12:01:00 AM"
            else:
                raise InvalidFormat(
                    "Error in Schedule: must supply at one of (30, 15, 5, 1) for every_x_minutes")
        else:
            raise InvalidFormat(
                "Error in Schedule: must supply one of (daily,hourly,on_receipt,days_of_week,days_of_month,every_x_minutes)")

        response = self._api.websdk.Config.Write.post(
            object_dn=work_dn,
            attribute_data=self._name_value_list(attributes, True)
        )
        response.assert_valid_response()

    def create(self, name: str, log_threshold: str = ClientWorkAttributeValues.SSHRemediation.LogThreshold.info,
               attributes: dict = None, get_if_already_exists: bool = True):
        """
        Args:
            name: The name of the client work.
            log_threshold: Set the logging level (defaults to INFO)
            attributes: Additional attributes to apply to the object.
            get_if_already_exists: If the objects already exists, just return it as is.

        Returns:
            :ref:`config_object` of the client work.
        """
        work_attributes = {
            ClientAgentSSHDiscoveryWorkAttributes.created_by   : ClientWorkAttributeValues.SSHRemediation.CreatedBy.websdk,
            ClientAgentSSHDiscoveryWorkAttributes.interval     : 0,
            ClientAgentSSHDiscoveryWorkAttributes.log_threshold: log_threshold
        }

        if attributes:
            work_attributes.update(attributes)

        return self._config_create(
            name=name,
            parent_folder_dn=self._work_base_dn,
            config_class=ClientAgentSSHProvisioningWorkAttributes.__config_class__,
            attributes=work_attributes,
            get_if_already_exists=get_if_already_exists
        )

    def unschedule(self, work: 'Union[Config.Object, str]'):
        """
        Removes any scheduling for the client work, but does not delete the client work.

        Args:
            work: :ref:`config_object` or name of the client work.
        """
        work_dn = self._get_dn(work, parent_dn=self._work_base_dn)
        for attribute_name in {
            ClientAgentSSHDiscoveryWorkAttributes.start_time,
            ClientAgentSSHDiscoveryWorkAttributes.schedule_type,
            ClientAgentSSHDiscoveryWorkAttributes.interval,
            ClientAgentSSHDiscoveryWorkAttributes.days_of_week,
            ClientAgentSSHDiscoveryWorkAttributes.days_of_month
        }:
            self._api.websdk.Config.ClearAttribute.post(
                object_dn=work_dn,
                attribute_name=attribute_name
            ).assert_valid_response()


@feature('User Certificate Creation')
class UserCertificateCreation(_ClientWorkBase):
    def __init__(self, api):
        super().__init__(api=api)

    def create(self, name: str, certificate_container: 'Union[Config.Object, str]', ca_template: 'Union[Config.Object, str]',
               contacts: 'List[Identity.Identity, str]', description: str = None,
               naming_pattern: str = ClientWorkAttributeValues.UserCertificateCreation.DefaultValues.naming_pattern,
               common_name: str = ClientWorkAttributeValues.UserCertificateCreation.DefaultValues.common_name,
               organization: str = ClientWorkAttributeValues.UserCertificateCreation.DefaultValues.organization,
               organizational_unit: List[
                   str] = ClientWorkAttributeValues.UserCertificateCreation.DefaultValues.organizational_unit,
               city_locality: str = ClientWorkAttributeValues.UserCertificateCreation.DefaultValues.city_locality,
               state_province: str = ClientWorkAttributeValues.UserCertificateCreation.DefaultValues.state_province,
               country: str = ClientWorkAttributeValues.UserCertificateCreation.DefaultValues.country,
               user_email: bool = False, subject_alt_names_email: bool = False, subject_alt_names_upn: bool = False,
               key_bit_strength: int = 2048, automatic_renewal: bool = True, renewal_days_before: int = 30,
               configure_outlook: bool = False, outlook_security_name: str = "", outlook_encrypt_messages: bool = False,
               outlook_send_cleartext_signed: bool = False, outlook_sign_outgoing: bool = False,
               outlook_request_receipts: bool = False,
               publish_to_identity_provider: bool = False, publish_pre_enrollment: bool = False,
               install_previous_certs: bool = False, allow_mobile_sharing: bool = False,
               lifecycle_groups: 'List[Union[Config.Object, str]]' = None, lifecycle_revoke_cert: bool = False,
               lifecycle_disable_cert: bool = False,
               portal_friendly_name: str = None, portal_icon: int = 0, portal_download_limit: int = 3,
               portal_instructions: str = None, attributes: dict = None, get_if_already_exists: bool = True):
        """
        Args:
            name: The name of the client work.
            certificate_container: :ref:`config_object` or :ref:`dn` of the folder to place certificates.
            ca_template: :ref:`config_object` or DN of the Certificate Authority.
            contacts: List of :ref:`identity_object` or :ref:`prefixed_name` of the contacts.)
            description: Description for the certificates
            naming_pattern: Object naming pattern
            common_name: Common name for the certificate
            organization: Organization for the certificate
            organizational_unit: A list of organizational units for the certificate
            city_locality: A city or locality for the certificate
            state_province: A state or province for the certificate
            country: A country code for the certificate
            user_email: Add user's email to the certificate
            subject_alt_names_email: Use subject alternative name email for the certificate
            subject_alt_names_upn: Use subject alternative upn for the certificate
            key_bit_strength: The key size of the certificate
            automatic_renewal: Enable automatic renewal for the certificate
            renewal_days_before: The number of days before expiration to renew the certificate
            configure_outlook: Configure Microsoft Outlook for Windows
            outlook_security_name: Security settings name
            outlook_encrypt_messages: Encrypt outgoing messages
            outlook_send_cleartext_signed: Send cleartext signed messages
            outlook_sign_outgoing: Sign outgoing messages
            outlook_request_receipts: Request S/MIME receipts
            publish_to_identity_provider: Publish to identity provider
            publish_pre_enrollment: Publish when a new identity is found (pre-enrollment)
            install_previous_certs: Install previous certificate versions
            allow_mobile_sharing: Allow certificate sharing with mobile devices
            lifecycle_groups: A list of :ref:`config_object` or :ref:`dn` of groups such that when a member is removed
                              from all groups in the list the certificate can be revoked and/or disabled
            lifecycle_revoke_cert: Revoke certificates when a user's membership is removed from all lifecycle groups
            lifecycle_disable_cert: Disable certificates when a user's membership is removed from all lifecycle groups
            portal_friendly_name: Portal friendly name
            portal_icon: Portal icon:
                                    0 - certificate
                                    1 - Envelope
                                    2 - Wi-Fi
                                    3 - VPN
            portal_download_limit: Limit the number of portal downloads
            portal_instructions: Text of portal download instructions
            attributes: Additional attributes to apply to the object.
            get_if_already_exists: If the objects already exists, just return it as is.

        Returns:
            :ref:`config_object` of the client work.
        """
        certificate_container_dn = self._get_dn(certificate_container)
        ca_template_dn = self._get_dn(ca_template)
        work_attributes = {
            ClientUserCertificateWorkAttributes.created_by                       : ClientWorkAttributeValues.UserCertificateCreation.CreatedBy.websdk,
            ClientUserCertificateWorkAttributes.certificate_container            : certificate_container_dn,
            ClientUserCertificateWorkAttributes.certificate_authority            : ca_template_dn,
            ClientUserCertificateWorkAttributes.naming_pattern                   : naming_pattern,
            ClientUserCertificateWorkAttributes.contact                          : [self._get_prefixed_universal(c) for c in contacts] if contacts else None,
            ClientUserCertificateWorkAttributes.key_bit_strength                 : key_bit_strength,
            ClientUserCertificateWorkAttributes.x509_subject                     : common_name,
            ClientUserCertificateWorkAttributes.organization                     : organization,
            ClientUserCertificateWorkAttributes.organizational_unit              : organizational_unit,
            ClientUserCertificateWorkAttributes.city                             : city_locality,
            ClientUserCertificateWorkAttributes.state                            : state_province,
            ClientUserCertificateWorkAttributes.country                          : country,
            ClientUserCertificateWorkAttributes.disable_automatic_renewal        : int(not automatic_renewal),
            ClientUserCertificateWorkAttributes.publish_to_identity              : int(publish_to_identity_provider),
            ClientUserCertificateWorkAttributes.publish_to_identity_on_pre_enroll: int(publish_pre_enrollment),
            ClientUserCertificateWorkAttributes.transfer_allowed                 : allow_mobile_sharing,
            ClientUserCertificateWorkAttributes.include_historic_certificates    : install_previous_certs,
            ClientUserCertificateWorkAttributes.certificate_icon                 : portal_icon,
            ClientUserCertificateWorkAttributes.download_limit                   : portal_download_limit,
        }

        if description: work_attributes[ClientUserCertificateWorkAttributes.description] = description
        if user_email:
            work_attributes[ClientUserCertificateWorkAttributes.x509_e] = \
                ClientWorkAttributeValues.UserCertificateCreation.DefaultValues.user_email
        if subject_alt_names_email:
            work_attributes[ClientUserCertificateWorkAttributes.x509_subjectaltname_rfc822] = \
                ClientWorkAttributeValues.UserCertificateCreation.DefaultValues.subject_alt_names_email
        if subject_alt_names_upn:
            work_attributes[ClientUserCertificateWorkAttributes.x509_subjectaltname_othername_upn] = \
                ClientWorkAttributeValues.UserCertificateCreation.DefaultValues.subject_alt_names_upn
        if automatic_renewal: work_attributes[ClientUserCertificateWorkAttributes.renewal_window] = renewal_days_before

        if configure_outlook:
            work_attributes[ClientUserCertificateWorkAttributes.outlook_profile_generation] = int(configure_outlook)
            work_attributes[ClientUserCertificateWorkAttributes.outlook_profile_name] = outlook_security_name
            option_value = 0
            if outlook_encrypt_messages: option_value += 1
            if outlook_sign_outgoing: option_value += 2
            if not outlook_send_cleartext_signed: option_value += 32
            if outlook_request_receipts: option_value += 512
            work_attributes[ClientUserCertificateWorkAttributes.outlook_profile_options] = option_value

        if lifecycle_groups:
            work_attributes[ClientUserCertificateWorkAttributes.required_member_identity] = [
                group.guid for group in lifecycle_groups
            ]
            work_attributes[ClientUserCertificateWorkAttributes.membership_loss_disable] = lifecycle_disable_cert
            work_attributes[ClientUserCertificateWorkAttributes.membership_loss_revoke] = lifecycle_revoke_cert

        if portal_friendly_name:
            work_attributes[ClientUserCertificateWorkAttributes.portal_friendly_name] = portal_friendly_name
        if portal_instructions:
            work_attributes[ClientUserCertificateWorkAttributes.download_instructions] = portal_instructions

        if attributes:
            work_attributes.update(attributes)

        return self._config_create(
            name=name,
            parent_folder_dn=self._work_base_dn,
            config_class=ClientUserCertificateWorkAttributes.__config_class__,
            attributes=work_attributes,
            keep_list_values=True,
            get_if_already_exists=get_if_already_exists
        )



from dataclasses import dataclass
from typing import List


@dataclass
class PKI:
    certificate_dn: str
    certificate_guid: str
    pki_dn: str
    pki_guid: str


@dataclass
class Certificate:
    city: str
    common_name: str
    country: str
    key_algorithm: str
    key_bit_size: str
    organization: str
    organizational_units: List[str]
    sans: 'List[SANS]'
    state: str


@dataclass
class Installation:
    credential_dn: str
    host: str


@dataclass
class SANS:
    name: str
    typename: str



from pytpp.attributes._helper import IterableMeta, Attribute


class ConnectionBaseAttributes(metaclass=IterableMeta):
	__config_class__ = "Connection Base"
	concurrent_connection_limit = Attribute('Concurrent Connection Limit')
	connection_method = Attribute('Connection Method')
	credential = Attribute('Credential')
	enforce_known_host = Attribute('Enforce Known Host', min_version='15.2')
	global_sudo = Attribute('Global sudo', min_version='15.3')
	host = Attribute('Host')
	last_known_fingerprint = Attribute('Last Known Fingerprint', min_version='15.2')
	last_known_key_type = Attribute('Last Known Key Type', min_version='17.1')
	port = Attribute('Port')
	remote_server_type = Attribute('Remote Server Type')
	secondary_credential = Attribute('Secondary Credential', min_version='15.3')
	sudo_password_delay = Attribute('Sudo Password Delay', min_version='17.1')
	temp_directory = Attribute('Temp Directory')
	terminal_columns = Attribute('Terminal Columns')
	terminal_rows = Attribute('Terminal Rows')
	terminal_type = Attribute('Terminal Type')
	timeout = Attribute('Timeout')
	trusted_fingerprint = Attribute('Trusted Fingerprint', min_version='15.2')
	trusted_key_type = Attribute('Trusted Key Type', min_version='17.1')



from pytpp.attributes._helper import IterableMeta, Attribute
from pytpp.attributes.application_base import ApplicationBaseAttributes


class VamSCA6000Attributes(ApplicationBaseAttributes, metaclass=IterableMeta):
	__config_class__ = "VamSCA6000"
	key_list_path = Attribute('Key List Path')
	sca6000_utility_path = Attribute('SCA6000 Utility Path')



from pytpp.properties.response_objects.dataclasses import processing_engines


class ProcessingEngines:
    @staticmethod
    def Engine(response_object: dict):
        if not isinstance(response_object, dict):
            response_object = {}
        return processing_engines.Engine(
            links=ProcessingEngines.Link(response_object.get('_links')),
            engine_dn=response_object.get('EngineDN'),
            engine_guid=response_object.get('EngineGuid'),
            engine_name=response_object.get('EngineName'),
        )

    @staticmethod
    def Folder(response_object: dict):
        if not isinstance(response_object, dict):
            response_object = {}
        return processing_engines.Folder(
            folder_dn=response_object.get('FolderDN'),
            folder_guid=response_object.get('FolderGuid'),
            folder_name=response_object.get('FolderName'),
        )

    @staticmethod
    def Link(response_object: dict):
        if not isinstance(response_object, dict):
            response_object = {}
        return processing_engines.Link(
            self=ProcessingEngines.Self(response_object.get('self')),
        )

    @staticmethod
    def Self(response_object: dict):
        if not isinstance(response_object, dict):
            response_object = {}
        return processing_engines.Self(
            href=response_object.get('href'),
        )



from pytpp.attributes._helper import IterableMeta
from pytpp.attributes.branch_base import BranchBaseAttributes


class SSOAttributes(BranchBaseAttributes, metaclass=IterableMeta):
	__config_class__ = "SSO"



from pytpp.attributes._helper import IterableMeta, Attribute
from pytpp.attributes.code_signing_environment_template_base import CodeSigningEnvironmentTemplateBaseAttributes


class CodeSigningAppleEnvironmentTemplateAttributes(CodeSigningEnvironmentTemplateBaseAttributes, metaclass=IterableMeta):
	__config_class__ = "Code Signing Apple Environment Template"
	cn_pattern = Attribute('CN Pattern', min_version='21.2')
	certificate_container_dn = Attribute('Certificate Container DN', min_version='21.2')
	key_container_dn = Attribute('Key Container DN', min_version='21.2')
	key_storage_location = Attribute('Key Storage Location', min_version='21.2')
	max_uses = Attribute('Max Uses', min_version='21.2')
	validity_period = Attribute('Validity Period', min_version='21.2')



from pytpp.attributes._helper import IterableMeta, Attribute
from pytpp.attributes.credential_base import CredentialBaseAttributes


class AmazonCredentialAttributes(CredentialBaseAttributes, metaclass=IterableMeta):
	__config_class__ = "Amazon Credential"
	authentication_credential = Attribute('Authentication Credential', min_version='18.3')
	authentication_source = Attribute('Authentication Source', min_version='18.3')
	ec2_assigned_role = Attribute('EC2 Assigned Role', min_version='21.2')
	region_code = Attribute('Region Code', min_version='18.3')
	role = Attribute('Role', min_version='18.3')
	web_service_url = Attribute('Web Service URL', min_version='18.3')



class ResultCodes:
    Client = {
        0 : 'Success, the operation succeeded, however, the client entry is untrusted',
        1 : 'SuccessButUntrusted',
        2 : 'ArgumentInvalidValue',
        3 : 'ArgumentUnsupportedDerivingClass',
        4 : 'ArgumentIsEmpty',
        5 : 'ArgumentIsNull',
        6 : 'BadAuthentication, the authentication check failed',
        7 : 'BeginTransactionFailed',
        8 : 'CommitTransactionFailed',
        9 : 'ConnectionError',
        10: 'ConfigObjectDoesNotExist',
        11: 'ConfigOperationFailed',
        12: 'EntryDoesNotExist',
        13: 'EntryAlreadyExists',
        14: 'EntryIsDisabled, the entry is disabled',
        15: 'EntryIsUntrusted, the entry is untrusted',
        16: 'Exception',
        17: 'DuplicateDriverRegistration, an existing storage driver has already been registered with a matching name',
        18: 'DuplicateIdentifer',
        19: 'DuplicateNetInterface',
        20: 'DuplicateOperatingSystemInTable',
        21: 'DuplicateManufacturerInTable',
        22: 'IdentityFromProviderIdFailed',
        23: 'InsufficientRights (also known as permissions)',
        24: 'InvalidCallingAssembly',
        25: 'InvalidCharacterInDistinguishedName',
        26: 'InvalidDistinguishedName',
        27: 'InvalidDriverName',
        28: 'InvalidDriverPrefix',
        29: 'InvalidKey (ClientId is not formatted correctly)',
        30: 'InvalidKeyRole (ClientId has an invalid role)',
        31: 'InvalidKeyId (ClientId is invalid)',
        32: 'InvalidRole',
        33: 'InvalidOperation',
        34: 'MachineAlreadyExists',
        35: 'MissingDriverInterface',
        36: 'MissingProviderSid',
        37: 'NoDriver',
        38: 'NoPendingOperations',
        39: 'NotImplemented',
        40: 'NotSupported',
        41: 'RightsOperationFailed',
        42: 'StatementExecutionFailed',
        43: 'TransactionInProgress',
        44: 'TryLater (Requested data was not yet available, retry later)',
        45: 'RemoteError (remote request failed; see Error property for details)',
        46: 'FailedToRetrieveSession (Session was not available)'
    }

    CodeSign = {
        0   : 'None	(Undefined result)',
        1   : 'Success	(The operation was successful)',
        2   : 'InsufficientPermission	(Insufficient permission to perform the operation)',
        3   : 'CreateFailure	(Failed to create an object)',
        4   : 'UpdateFailure	(Failed to update an object)',
        5   : 'DeleteFailure	(Failed to delete an object)',
        6   : 'RenameFailure	(Failed to rename an object)',
        7   : 'ObjectNotFound	(Expected one object type but was requesting another)',
        8   : 'IncorrectObjectType	(Operation prohibited on objects with children)',
        9   : 'ObjectHasChildren	(Failed to assign permissions)',
        10  : 'AssignPermissionFailure	(Object was saved and related objects had changes that were not saved)',
        11  : 'RelatedObjectNotSaved	(An error occurred when trying to execute the command via REST API)',
        12  : 'RemoteError	(The requested functionality has not been implemented yet)',
        1000: 'AssignProjectPermissionFailure	(Project may have been saved, but assigning role permissions failed)',
        1001: 'DuplicateUserAssignment	(The user/group was added to a role they were already assigned to)',
        1002: 'UserNotFound	(The user/group was not found in the role)',
        2000: 'ObjectInGroup	(Tried to add an object to the group that was already there)',
        2001: 'ObjectNotInGroup	(Tried to remove an object from the group that was not in the group)',
        2002: 'ObjectAlreadyExists	(Tried to create an object and an object with the same name already exists)',
        3000: 'EnvironmentMoveRestricted	(Tried to change the project containing an environment)',
        3001: 'ValueProhibitedByTemplate	(An attribute on the environment does not match what is required by the template)',
        3002: 'KeyAccessFailure	(An attempt to access the key in this environment has failed)',
        3003: 'QueryFailure	(An error occurred while performing a search)',
        3004: 'ImportFailure	(An error occurred while performing an import)',
        3005: 'ObjectsPending	(The key, certificate, or other object is not yet ready for this environment)',
    }

    Config = {
        1  : 'Success',
        2  : 'InvalidArgument',
        3  : 'InvalidArgumentRange',
        4  : 'MismatchedArguments',
        5  : 'NotImplemented',
        6  : 'InvalidDestinationList',
        7  : 'InsufficientPrivileges',
        8  : 'InvalidOperation',
        9  : 'UnexpectedAssemblyError',
        10 : 'OutOfMemory',
        100: 'AttributeDoesNotExist',
        101: 'AttributeAlreadyExists',
        102: 'AttributeNotFound',
        103: 'AttributeValueExists',
        104: 'AttributeStillInUse',
        105: 'AttributeNameTooLong',
        106: 'AttributeReferenceDoesNotExist',
        107: 'AttributeSyntaxCollision',
        108: 'AttributePropertyCollision',
        109: 'CannotRemoveMandatory',
        110: 'AttributeValueIsMandatory',
        111: 'AttributeValueTooLong',
        112: 'IllegalAttributeForClass',
        113: 'InvalidAttributeDN',
        114: 'AttributeValueDoesNotExist',
        115: 'AttributeIsSingleValued',
        116: 'AttributeIsReadOnly',
        117: 'AttributeIsHidden',
        200: 'ClassDoesNotExist',
        201: 'ClassAlreadyExists',
        202: 'ClassStillInUse',
        203: 'ClassNameTooLong',
        204: 'ClassInvalidSuperClass',
        205: 'ClassInvalidContainmentClass',
        206: 'ClassInvalidNamingAttribute',
        207: 'ClassInvalidMandatoryAttribute',
        208: 'ClassInvalidOptionalAttribute',
        209: 'ClassInvalidName',
        210: 'ClassInvalidContainmentSubClass',
        211: 'ClassInvalidPolicyClass',
        300: 'PolicyDoesNotExist',
        301: 'PolicyLockStateCollision',
        350: 'LockNameAlreadyExists',
        351: 'LockNameDoesNotExist',
        352: 'LockNameOwnedByAnother',
        353: 'LockNameLimitReached',
        354: 'LockNameAttemptTimedOut',
        400: 'ObjectDoesNotExist',
        401: 'ObjectAlreadyExists',
        402: 'ObjectHasChildren',
        403: 'ObjectNameTooLong',
        404: 'ObjectDepthTooDeep',
        405: 'ObjectInvalidName',
        406: 'ObjectInvalidClass',
        407: 'ObjectInvalidContainment',
        408: 'ObjectMandatoryMissing',
        409: 'ObjectIsReadOnly',
        410: 'ObjectInvalidOperation',
        500: 'DriverMissingDSN',
        501: 'DriverMissingDatabaseName',
        502: 'DriverDatabaseError',
        503: 'DriverTransactionError',
        504: 'DriverTransactionCollision',
        505: 'DriverGenerationUpdateError',
        600: 'CacheLockException',
        601: 'CacheEntryNotFound',
        602: 'CacheEntryAlreadyExists',
        603: 'CacheEntryIsSuperior',
        604: 'CacheEntryIsIncompatible',
        700: 'XmlInvalidStructure',
        701: 'XmlMissingNaming',
        702: 'XmlMissingSyntax',
        703: 'XmlMissingProperty',
        704: 'XmlUnknownElementAttribute'
    }

    Credential = {
        1   : 'Success',
        2   : 'InvalidArgument',
        3   : 'InvalidArgumentRange',
        4   : 'MismatchedArguments',
        5   : 'NotImplemented',
        6   : 'InvalidDestinationList',
        7   : 'InsufficientPrivileges',
        8   : 'InvalidOperation',
        9   : 'UnexpectedAssemblyError',
        10  : 'OutOfMemory',
        100 : 'AttributeDoesNotExist',
        101 : 'AttributeAlreadyExists',
        102 : 'AttributeNotFound',
        103 : 'AttributeValueExists',
        104 : 'AttributeStillInUse',
        105 : 'AttributeNameTooLong',
        106 : 'AttributeReferenceDoesNotExist',
        107 : 'AttributeSyntaxCollision',
        108 : 'AttributePropertyCollision',
        109 : 'CannotRemoveMandatory',
        110 : 'AttributeValueIsMandatory',
        111 : 'AttributeValueTooLong',
        112 : 'IllegalAttributeForClass',
        113 : 'InvalidAttributeDN',
        114 : 'AttributeValueDoesNotExist',
        115 : 'AttributeIsSingleValued',
        116 : 'AttributeIsReadOnly',
        117 : 'AttributeIsHidden',
        200 : 'ClassDoesNotExist',
        201 : 'ClassAlreadyExists',
        202 : 'ClassStillInUse',
        203 : 'ClassNameTooLong',
        204 : 'ClassInvalidSuperClass',
        205 : 'ClassInvalidContainmentClass',
        206 : 'ClassInvalidNamingAttribute',
        207 : 'ClassInvalidMandatoryAttribute',
        208 : 'ClassInvalidOptionalAttribute',
        209 : 'ClassInvalidName',
        210 : 'ClassInvalidContainmentSubClass',
        300 : 'PolicyDoesNotExist',
        301 : 'PolicyLockStateCollision',
        350 : 'LockNameAlreadyExists',
        351 : 'LockNameDoesNotExist',
        352 : 'LockNameOwnedByAnother',
        353 : 'LockNameLimitReached',
        354 : 'LockNameAttemptTimedOut',
        400 : 'Disadvantageous',
        401 : 'ObjectAlreadyExists',
        402 : 'ObjectHasChildren',
        403 : 'ObjectNameTooLong',
        404 : 'ObjectDepthTooDeep',
        405 : 'ObjectInvalidName',
        406 : 'ObjectInvalidClass',
        407 : 'ObjectInvalidContainment',
        408 : 'ObjectMandatoryMissing',
        409 : 'ObjectIsReadOnly',
        410 : 'ObjectInvalidOperation',
        500 : 'DriverMissingDSN',
        501 : 'DriverMissingDatabaseName',
        502 : 'DriverDatabaseError',
        503 : 'DriverTransactionError',
        504 : 'DriverTransactionCollision',
        505 : 'DriverGenerationUpdateError',
        600 : 'CacheLockException',
        601 : 'CacheEntryNotFound',
        602 : 'CacheEntryAlreadyExists',
        603 : 'CacheEntryIsSuperior',
        604 : 'CacheEntryIsIncompatible',
        700 : 'XmlInvalidStructure',
        701 : 'XmlMissingNaming',
        702 : 'XmlMissingSyntax',
        703 : 'XmlMissingProperty',
        704 : 'XmlUnknownElementAttribute',
        1000: 'SecretStoreFailed',
        1001: 'AddAttributeFailed',
        1002: 'UnexpectedException',
        1003: 'PartialDeleteFailure',
        1004: 'CredentialTypeMismatch',
        1005: 'NoDriver',
        1006: 'VaultTypeMismatch',
        1007: 'DriverDenied',
        1008: 'VaultDataUnrecognized'
    }

    CertificatesWorkToDo = {
        0  : 'Success',
        1  : 'MissingDSN',
        2  : 'MissingDatabaseName',
        99 : 'UnexpectedAssemblyError',
        100: 'DatabaseError',
        101: 'DatabaseConnectionUnavailable',
        102: 'TransactionError',
        103: 'TransactionCollision',
        200: 'InsufficientPrivileges',
        300: 'IdentifierDoesNotExist',
        301: 'IdentifierAlreadyExists'
    }

    ChainValidation = {
        0  : 'None',
        2  : 'Success',
        4  : 'NoChain',
        8  : 'InvalidChain',
        16 : 'ExpiringChain',
        32 : 'IncompleteChain',
        64 : 'BlacklistedChain',
        128: 'MismatchedChain',
        512: 'MismatchedCertificate'
    }

    Discovery = {
        0  : 'None',
        1  : 'Success',
        2  : 'InvalidArgument',
        3  : 'InvalidArgumentRange',
        4  : 'MismatchedArguments',
        5  : 'NotImplemented',
        6  : 'InvalidDestinationList',
        7  : 'InsufficientPrivileges',
        8  : 'UnexpectedAssemblyError',
        99 : 'ConfigAssemblyError',
        100: 'InvalidDiscoveryDefinition',
        102: 'PendingPreviousRun',
        500: 'DriverMissingDSN',
        501: 'DriverMissingDatabaseName',
        502: 'DriverDatabaseError',
        503: 'DriverTransactionError',
        504: 'DriverTransactionCollision',
        505: 'DriverGenerationUpdateError',
        506: 'DriverConnectionError',
        507: 'DriverUnexpectedQueryResult',
        600: 'RemoveVaultEntryError',
    }

    EndEntityValidation = {
        -2147483648: 'UnexpectedError',
        0          : 'None',
        2          : 'Success',
        4          : 'HostResolutionFailed',
        8          : 'SettingError',
        16         : 'ConnectionFailure',
        32         : 'NoCertificateFound',
        64         : 'MismatchWithPrevious',
        128        : 'MismatchWithKnown',
        256        : 'MismatchWithUnknown',
        512        : 'NoLocalCertificate',
        536870912  : 'NotSupported',
        1073741824 : 'RetryLater'
    }

    Flow = {
        0 : 'Success',
        1 : 'InsufficientPermission',
        2 : 'TicketNotFound',
        3 : 'FailedToEnumerateTickets',
        4 : 'FailedToApproveTickets',
        5 : 'FailedToRejectTickets',
        6 : 'FailedToCreateTicket',
        7 : 'FailedToGetTicket',
        8 : 'FailedToDeleteTicket',
        9 : 'FailedToUseTicket',
        10: 'NoTicketUsesRemain',
        11: 'FailedToUpdateTickets'
    }

    Metadata = {
        0 : 'Success',
        1 : 'InvalidConfigObject',
        2 : 'InvalidDN',
        3 : 'InvalidName',
        4 : 'InvalidItem',
        5 : 'InvalidClass',
        6 : 'InvalidMetadataObject',
        7 : 'InvalidRights',
        8 : 'ItemIsNull',
        9 : 'ItemAlreadyExists',
        10: 'ItemTypeUnknown',
        11: 'ConfigCreateFailed',
        12: 'ConfigWriteFailed',
        13: 'ConfigDeleteFailed',
        14: 'MetadataInUse',
        15: 'NoAllowedValues',
        16: 'AllowedValueDoesNotExist',
        17: 'ValueNotInAllowedList',
        18: 'ItemNotValidForClass',
        19: 'NameTooLong',
        20: 'TooManyContainers',
        21: 'ConfigDnNotContainer',
        22: 'InvalidPolicyState',
        23: 'ConfigLockFailed',
        24: 'ConfigReadFailed',
        25: 'RemoteError'
    }

    SecretStore = {
        0 : 'Success',
        1 : 'InvalidCallingAssembly',
        2 : 'CreateDatabaseError',
        3 : 'UseDatabaseError',
        4 : 'CreateTableError',
        5 : 'CreateIndexError',
        6 : 'ConnectionError',
        7 : 'TransactionError',
        8 : 'InvalidVaultID',
        9 : 'InvalidParams',
        10: 'InsufficientPermissions',
        11: 'CryptoFailure',
        12: 'DeleteSecretFailed',
        13: 'AddSecretFailed',
        14: 'RetrieveSecretFailed',
        15: 'RetrieveSecretTypeFailed',
        16: 'GetNextVaultIDFailed',
        17: 'DisassociateFailed',
        18: 'OwnerLookupFailed',
        19: 'AssociateDataFailed',
        20: 'LookupFailed',
        21: 'InvalidKey',
        22: 'QueryError',
        23: 'SecurityGroupNotImplemented'
    }

    SSH = {
        1 : 'IsRootAccessOrphan',
        2 : 'IsUserAccessOrphan',
        3 : 'IsPrivateKeyOrphan',
        4 : 'IsKnownHostOrphan',
        5 : 'IsRootAccess',
        6 : 'IsDuplicatePrivateKey',
        7 : 'IsIllegalAlgorithm',
        8 : 'IsIllegalProtocolVersion',
        9 : 'IsIllegalVendorFormat',
        10: 'IsIllegalForcedCommand',
        11: 'IsIllegalSourceRestrictions',
        12: 'IsMissingOptions',
        13: 'IsKeyOlderThanAllowed',
        14: 'IsKeySmallerThanRequired',
        15: 'IsKeyLengthSmallerThan768',
        16: 'IsSharedPrivateKey',
        17: 'Not Currently in Use	',
        18: 'IsSharedServerAccount',
        19: 'IsPassphraseUnknown',
        20: 'IsUnknownClient',
        21: 'KeyWasNotRotated',
    }

    SSHErrorCodes = {
        1  : 'KeyNotFound',
        2  : 'KeysetNotFound',
        3  : 'DeviceNotFound',
        4  : 'BadKeyFormat',
        5  : 'KeyAlreadyExists',
        6  : 'NoPrivateKeyOnTpp',
        7  : 'UnsupportedKeyLength',
        8  : 'OperationIsAlreadyRunning',
        9  : 'NotAbleToLoadApprovers',
        10 : 'TicketNotFound',
        11 : 'MissingRejectComment',
        12 : 'InsufficientRights',
        13 : 'FolderNotFound',
        14 : 'KeyCannotBeDownloaded',
        15 : 'KeyHasNoMaterial',
        16 : 'InvalidOwnerIdentity',
        17 : 'FolderHasNoRights',
        18 : 'ImproperKeysetStage',
        19 : 'InappropriateTrustType "Key has inappropriate trust type"',
        100: 'InternalError',
    }

    ToDo = {
        0  : 'Success',
        1  : 'MissingDSN',
        2  : 'MissingDatabaseName',
        99 : 'UnexpectedAssemblyError',
        100: 'DatabaseError',
        101: 'DatabaseConnectionUnavailable',
        102: 'TransactionError',
        103: 'TransactionCollision',
        200: 'InsufficientPrivileges',
        300: 'IdentifierDoesNotExist',
        301: 'IdentifierAlreadyExists',
    }

    ValidationProtocols = {
        12  : "SSL 2.0",
        48  : "SSL 3.0",
        60  : "SSL 2.0, SSL 3.0",
        192 : "TLS 1.0",
        204 : "SSL 2.0, TLS 1.0",
        240 : "SSL 3.0, TLS 1.0",
        252 : "SSL 2.0, SSL 3.0, TLS 1.0",
        768 : "TLS 1.1",
        780 : "SSL 2.0, TLS 1.1",
        816 : "SSL 3.0, TLS 1.1",
        960 : "TLS 1.0, TLS 1.1",
        828 : "SSL 2.0, SSL 3.0, TLS 1.1",
        972 : "SSL 2.0, TLS 1.0, TLS 1.1",
        1008: "SSL 3.0, TLS 1.0, TLS 1.1",
        1020: "SSL 2.0, SSL 3.0, TLS 1.0, TLS 1.1",
        3072: "TLS 1.2",
        3084: "SSL 2.0, TLS 1.2",
        3120: "SSL 3.0, TLS 1.2",
        3264: "TLS 1.0, TLS 1.2",
        3840: "TLS 1.1, TLS 1.2",
        3132: "SSL 2.0, SSL 3.0, TLS 1.2",
        3276: "SSL 2.0, TLS 1.0, TLS 1.2",
        3852: "SSL 2.0, TLS 1.1, TLS 1.2",
        3312: "SSL 3.0, TLS 1.0, TLS 1.2",
        3888: "SSL 3.0, TLS 1.1, TLS 1.2",
        4032: "TLS 1.0, TLS 1.1, TLS 1.2",
        3324: "SSL 2.0, SSL 3.0, TLS 1.0, TLS 1.2",
        3900: "SSL 2.0, SSL 3.0, TLS 1.1, TLS 1.2",
        4044: "SSL 2.0, TLS 1.0, TLS 1.1, TLS 1.2",
        4080: "SSL 3.0, TLS 1.0, TLS 1.1, TLS 1.2",
        4092: "SSL 2.0, SSL 3.0, TLS 1.0, TLS 1.1, TLS 1.2"
    }

    Workflow = {
        0: 'None',
        1: 'Success',
        2: 'GenericFailure',
        3: 'TicketDoesNotExist',
        4: 'InsufficientPrivileges',
        5: 'BadArguments',
        6: 'ObjectDoesNotExist',
        7: 'RemoteError',
        8: 'WorkflowObjectDoesNotExist',
    }



from pytpp.attributes._helper import IterableMeta, Attribute
from pytpp.attributes.application_base import ApplicationBaseAttributes


class HashiCorpVaultPKIAttributes(ApplicationBaseAttributes, metaclass=IterableMeta):
	__config_class__ = "HashiCorp Vault PKI"
	crl_address = Attribute('CRL Address', min_version='20.1')
	create_certificate_authority = Attribute('Create Certificate Authority', min_version='20.1')
	create_pki_role = Attribute('Create PKI Role', min_version='20.1')
	enhanced_key_usage = Attribute('Enhanced Key Usage', min_version='20.1')
	network_validation_disabled = Attribute('Network Validation Disabled', min_version='20.1')
	ocsp_address = Attribute('OCSP Address', min_version='20.1')
	policydn = Attribute('PolicyDN', min_version='20.1')
	role_name = Attribute('Role Name', min_version='20.1')



from dataclasses import dataclass


@dataclass
class Engine:
    links: 'Link'
    engine_dn: str
    engine_guid: str
    engine_name: str


@dataclass
class Folder:
    folder_dn: str
    folder_guid: str
    folder_name: str


@dataclass
class Link:
    self: 'Self'


@dataclass
class Self:
    href: str



from pytpp.attributes._helper import IterableMeta, Attribute
from pytpp.attributes.device import DeviceAttributes


class SSHManualInstallKeyDeviceAttributes(DeviceAttributes, metaclass=IterableMeta):
	__config_class__ = "SSH Manual Install Key Device"
	ssh_manual_install_key_contact_email = Attribute('SSH Manual Install Key Contact Email', min_version='18.3')
	ssh_manual_install_key_location = Attribute('SSH Manual Install Key Location', min_version='18.3')
	ssh_manual_install_key_notes = Attribute('SSH Manual Install Key Notes', min_version='18.3')
	ssh_manual_install_key_owner = Attribute('SSH Manual Install Key Owner', min_version='18.3')



from pytpp.attributes._helper import IterableMeta, Attribute
from pytpp.attributes.service_module import ServiceModuleAttributes


class ValidationManagerAttributes(ServiceModuleAttributes, metaclass=IterableMeta):
	__config_class__ = "Validation Manager"
	start_time = Attribute('Start Time')
